#!/bin/bash

#############
# VARIABLES #
#############
MAXIMUM_ADML_ASYNC_RATE=50
MAXIMUM_SUGGESTED_CLIENT_CONNECTION_PER_ADML_INSTANCE=10
ADML_INSTANCES__ST_CONF_FILE=.st_conf_adml_instances
ADML_CONNECTIONS__ST_CONF_FILE=.st_conf_adml_connections
CYCLE_REPEATS__ST_CONF_FILE=.st_conf_cycle_repeats
N_TESTCASES__ST_CONF_FILE=.st_conf_n_testcases
ADML_RATE_PER_INSTANCE__ST_CONF_FILE=.st_conf_rate_per_instance
ADML_DESIRED_RATE__ST_CONF_FILE=.st_conf_desired_rate
N_TESTCASES_PROGRAM_LAYOUT__ST_CONF_FILE=.st_conf_n_testcases_program_layout

#############
# FUNCTIONS #
#############
_exit () {
  echo
  echo $1
  echo
  exit 1
}

# ceil of division $1/$2
ceil() {
  #echo "$1 $2" | awk '{print int( ($1/$2) + 1 )}'
  awk -vnumber="$1" -vdiv="$2" '
  function ceiling(x){return x%1 ? int(x)+1 : x}
  BEGIN{ print ceiling(number/div) }'
}

# Calculates the number of ADML instances and their client connections
calculate_deployment_layout() {
  echo "Input the maximum client connections accepted by the server to be tested [5000]:"
  read max_server_accepted_connections
  [ "$max_server_accepted_connections" = "" ] && max_server_accepted_connections=5000

  echo "Input the maximum desired test case rate per second:"
  read desired_rate
  while [ -z "$desired_rate" ]; do read desired_rate; done

  max_connections=$((desired_rate/MAXIMUM_ADML_ASYNC_RATE))
  if [ $max_connections -eq 0 ]
  then
    G_ADML_CONNECTIONS=1
    G_ADML_INSTANCES=1
    return
  elif [ $max_connections -gt $max_server_accepted_connections ]
  then
    _exit "Not enough server connections to fit the desired rate (requires $max_connections connections)."
  fi

  echo
  echo "==========================================================================================================="
  echo "Orientative table"
  echo "-----------------------------------------------------------------------------------------------------------"
  echo -n "Number of instances:  "
  instances__dflt=$(ceil $max_connections $MAXIMUM_SUGGESTED_CLIENT_CONNECTION_PER_ADML_INSTANCE)
  for conn in `seq 1 $MAXIMUM_SUGGESTED_CLIENT_CONNECTION_PER_ADML_INSTANCE | tac`
  do
    instances=$(ceil $max_connections $conn)
    echo -n -e "\t$instances"
    [ $conn -eq 1 ] && instances__dflt=$instances
  done
  echo
  echo -e "Connects per instance:\t10\t9\t8\t7\t6\t5\t4\t3\t2\t1"
  echo "==========================================================================================================="
  echo
  #echo "Input selection (connections per instance 1..$MAXIMUM_SUGGESTED_CLIENT_CONNECTION_PER_ADML_INSTANCE) [1]:"
  echo "Input the desired number of ADML instances [$instances__dflt]:"
  echo " (more than $instances instances, implies 1 single connection/instance)"
  read G_ADML_INSTANCES
  [ -z "$G_ADML_INSTANCES" ] && G_ADML_INSTANCES=$instances__dflt
  [ $G_ADML_INSTANCES -lt 1 ] && G_ADML_INSTANCES=1
  G_ADML_CONNECTIONS=$(ceil $max_connections $G_ADML_INSTANCES)
  if [ $G_ADML_CONNECTIONS -gt $MAXIMUM_SUGGESTED_CLIENT_CONNECTION_PER_ADML_INSTANCE ]
  then
    echo "Warning: the number of connections per ADML instance ($G_ADML_CONNECTIONS) is greater"
    echo "         than the maximum recommended: $MAXIMUM_SUGGESTED_CLIENT_CONNECTION_PER_ADML_INSTANCE"
    echo
    echo "Press ENTER to continue, CTRL-C to abort ..."
    read dummy
  fi
  client_connections=$((G_ADML_INSTANCES*G_ADML_CONNECTIONS))
  if [ $client_connections -gt $max_server_accepted_connections ]
  then
    echo
    echo "Insufficient server connections available ($max_server_accepted_connections) to accept"
    echo " launcher client connections: $G_ADML_INSTANCES x $G_ADML_CONNECTIONS = $client_connections."
    _exit "Configuration error"
  fi
}

#############
# EXECUTION #
#############

cd `dirname $0`

echo
echo "====================================="
echo "ADML SYSTEM TEST CONFIGURATION WIZARD"
echo "====================================="
echo
if [ -d ADMLS ]
then
  echo
  echo "Detected configured layout (*):"
  echo
  ADML_INSTANCES=`cat $ADML_INSTANCES__ST_CONF_FILE 2>/dev/null`
  ADML_CONNECTIONS=`cat $ADML_CONNECTIONS__ST_CONF_FILE 2>/dev/null`
  RATE_PER_INSTANCE=`cat $ADML_RATE_PER_INSTANCE__ST_CONF_FILE 2>/dev/null`
  DESIRED_RATE=`cat $ADML_DESIRED_RATE__ST_CONF_FILE 2>/dev/null`
  REPEATS=`cat $CYCLE_REPEATS__ST_CONF_FILE 2>/dev/null`
  N_TESTCASES=`cat $N_TESTCASES__ST_CONF_FILE 2>/dev/null`

  [ -n "$ADML_INSTANCES" ] && echo "  $ADML_INSTANCES ADML instances."
  [ -n "$ADML_CONNECTIONS" ] && echo "  $ADML_CONNECTIONS connections per ADML instance."
  [ -n "$RATE_PER_INSTANCE" ] && echo "  Rate of $RATE_PER_INSTANCE TCs/second and instance."
  [ -n "$DESIRED_RATE" ] && echo "  Desired rate was: $DESIRED_RATE TCs/second and instance."
  [ -n "$REPEATS" ] && echo "  Number of cycles: $((REPEATS + 1))."
  [ -n "$N_TESTCASES" ] && echo "  $N_TESTCASES test cases programmed."
  echo
  operation_result=$(echo "scale=3 ; (1 + $REPEATS) * $N_TESTCASES / $ADML_INSTANCES / $RATE_PER_INSTANCE" | bc)
  echo "  Time covered: $((REPEATS + 1)) cycles x ($N_TESTCASES/(${ADML_INSTANCES}x${RATE_PER_INSTANCE})) = $operation_result seconds"
  echo

  echo "  (*) As ADMLS directory still exists you should move/remove it to continue."
  echo "      You may have to 'pkill ADML' before (resources busy)."
  _exit
fi

[ ! -d services ] && _exit "Missing services configuration (expecting '$PWD/services' directory) !"

calculate_deployment_layout

# Format for instance number (for example: 1 -> 001, .., 45 -> 045, 300 -> 300)
INSTANCE_FORMAT=$(echo $G_ADML_INSTANCES | wc -c)
INSTANCE_FORMAT=$((INSTANCE_FORMAT-1))

# Dump persintently:
echo $G_ADML_INSTANCES > $ADML_INSTANCES__ST_CONF_FILE
echo $G_ADML_CONNECTIONS > $ADML_CONNECTIONS__ST_CONF_FILE

# Rate per instance:
rate_per_instance=$((desired_rate/$G_ADML_INSTANCES))
[ $rate_per_instance -lt 1 ] && rate_per_instance=1
echo $rate_per_instance > $ADML_RATE_PER_INSTANCE__ST_CONF_FILE
echo $desired_rate > $ADML_DESIRED_RATE__ST_CONF_FILE

echo
echo "Layout:"
echo
echo " - $G_ADML_INSTANCES ADML instances"
echo " - $G_ADML_CONNECTIONS client connections per ADML instance"
maximum_rate=$((G_ADML_INSTANCES*G_ADML_CONNECTIONS*MAXIMUM_ADML_ASYNC_RATE))
overcommit_rate_per_instance=$((G_ADML_CONNECTIONS*MAXIMUM_ADML_ASYNC_RATE))
echo
echo " - Desired rate: $desired_rate test cases per second:"
echo "   Remember the command to send the needed rate per instance ($rate_per_instance TCs/sec):"
echo "       ./operation.sh \"test|ttps|$rate_per_instance\""
echo
if [ $maximum_rate -ne $desired_rate ]
then
  echo " - Maximum bunch rate ($G_ADML_CONNECTIONS connections per instance): $maximum_rate testcases/second"
  echo "   For this, configure instances rate to $overcommit_rate_per_instance test cases per second:"
  echo "       ./operation.sh \"test|ttps|$overcommit_rate_per_instance\""
  echo
fi
echo "Usually, you will program a test case per subscriber."
echo "Input the number of test cases to program:"
read N_TESTCASES
while [ -z "$N_TESTCASES" ]; do read N_TESTCASES; done
echo $N_TESTCASES > $N_TESTCASES__ST_CONF_FILE
echo "Input the first test id to program [1]:"
read first_value
[ "$first_value" = "" ] && first_value=1
[ $first_value -lt 1 ] && first_value=1
echo
time_covered=$(echo "$N_TESTCASES/$desired_rate" | bc)
testcase_per_adml_instance=$((N_TESTCASES/G_ADML_INSTANCES))
echo "Aproximately, $testcase_per_adml_instance test cases will be programmed on each ADML instance (we have $G_ADML_INSTANCES instances)."
echo "As we shall program $N_TESTCASES test cases, the total time covered for the desired system rate of $desired_rate cps, is $time_covered seconds."

echo
if [ $time_covered -lt 10 ]
then
  echo "$time_covered seconds is under 10 seconds, you should add more test cases to the pool except if you are sure"
  echo " they will take less time that cycle completion. You could ensure that with a first timeout step."
  echo "Configuring such timeout slightly under $((1000*time_covered)) milliseconds, you could repeat the cycle safely to"
  echo " obtain a greater total time of testing."
fi
echo
echo "How many total time you need to cover (in minutes):"
read minutes
while [ -z "$minutes" ]; do read minutes; done
seconds=$((minutes*60))
repeats=0
[ $seconds -gt $time_covered ] && { repeats=$(ceil $seconds $time_covered) ; repeats=$((repeats-1)) ; }
if [ $repeats -gt 0 ]
then
  cycles=$((repeats+1))
  echo "You will need $cycles cycles to cover $minutes minutes."
  echo "Input the desired number of cycles [$cycles]: "
  echo " (providing 1, you will cover $time_covered seconds)"
  echo
  read wanted_cycles
  [ -z "$wanted_cycles" ] && wanted_cycles=$cycles
  wanted_repeats=$((wanted_cycles-1))
  echo $wanted_cycles > $CYCLE_REPEATS__ST_CONF_FILE
  [ $wanted_repeats -gt 0 ] && echo "Configured $wanted_repeats cycle repeats ($wanted_cycles cycles x $time_covered seconds ~ $((wanted_cycles*time_covered)) seconds of testing)."
else
  echo 0 > $CYCLE_REPEATS__ST_CONF_FILE
fi
echo
echo "System test configuration completed."
echo
echo "Ready to clone/start the ADML instances: press ENTER to continue, CTRL+C to abort ..."
read dummy

# Update services.xml regarding the number of client connections:
cd services 
cp services.msk services.xml
sed -i 's/__CLIENT_CONNECTIONS__/'$G_ADML_CONNECTIONS'/g' services.xml
cd - >/dev/null

# Create instances and layout:
0>$N_TESTCASES_PROGRAM_LAYOUT__ST_CONF_FILE
for i in `seq 1 $G_ADML_INSTANCES`
do
  instance=$(printf "%0${INSTANCE_FORMAT}d" $i)
  offset=$((testcase_per_adml_instance * (i-1)))
  ini=$((offset + first_value))
  fin=$((offset + first_value + testcase_per_adml_instance - 1))
  echo "$instance $ini $fin" >> $N_TESTCASES_PROGRAM_LAYOUT__ST_CONF_FILE

  echo "Creating ADML instance $instance ..."
  mkdir -p ADMLS/ADML-$instance
  cd ADMLS/ADML-$instance
  mkdir counters
  mkdir test-reports
  ln -s ../../dynlibs
  # Create resources:
  ln -s ../../.operation-one.sh operation.sh
  ln -s ../../pre-start.sh
  cp ../../.run-one.sh run.sh
  sed -i 's/^EXE=.*/EXE=ADML-'$instance'/' run.sh
  ln -s ../../ADML ADML-$instance
  for xml in `ls ../../services/*xml`; do ln -s $xml; done
  cd - >/dev/null
done

echo
echo "Now you can run all the instances deployed: ./run.sh"
echo "To configure another layout you should execute this script again."
echo
echo "Done!"

