#!/bin/bash

# ANNA - Anna is Not Nothingness Anymore
#
# (c) Copyright 2005-2014 Eduardo Ramos Testillano & Francisco Ruiz Rayo
#
# http://redmine.teslayout.com/projects/anna-suite
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#     * Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above
# copyright notice, this list of conditions and the following disclaimer
# in the documentation and/or other materials provided with the
# distribution.
#     *  Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# Authors: eduardo.ramos.testillano@gmail.com
#          cisco.tierra@gmail.com


# Decoder version using tshark tool
#
# RESTRICTIONS: TCP & SCTP datagrams. Allow split diameter messages between frames, but not, several messages into single datagram.
#               If sniffing fails, try to create the <frame>.hex and <frame>.metadata files manually (try 'tshark -nr input.pcap -Vx' and so on ...)

#############
# VARIABLES #
#############
tmpdir=$(mktemp -d)
TSHARK=tshark

#############
# FUNCTIONS #
#############

usage() {
  echo
  echo "Usage: $0 [-h|--help] [-o|--other-ports <list>] [-d|--results-dir <dir>] [-s|--sids <file>] <pcap>"
  echo
  echo "       -h|--help:                this usage help."
  echo "       -o|--other-ports <list>:  space-separated list of ports which frames"
  echo "                                 will be decoded as diameter protocol although"
  echo "                                 not being standard. For example, we could use"
  echo "                                 \"13868\" to disect the Ericsson Sy variant."
  echo "       -d|--results-dir <dir>:   directory where results are stored."
  echo "                                 By default, pcap dirname."
  echo "       -s|--sids <file>:         file containing a list of Session-Id values"
  echo "                                 (one per line) to be taken into account."
  echo "                                 If missing, all the frames will be extracted."
  echo
  echo "       pcap:                     pcap formatted file to be processed."
  echo
  echo "       The utility, dumps the extracted hexadecimal content and useful information"
  echo "       (timestamps, source, destination, etc.) within a metadata file:"
  echo
  echo "          <results directory>/<frame sequence>.hex"
  echo "          <results directory>/<frame sequence>.metadata"
  echo
  _exit
}

parse_arguments() {
  OTHER_PORTS=
  RESULTS_DIR=
  SIDS_FILE=
  PCAP_FILE=

  while [ $# -gt 0 ]; do
    case $1 in
      -h|--help)
        usage
      ;;

      -o|--other-ports)
        OTHER_PORTS="$2"
        [ -z "$OTHER_PORTS" ] && _exit "Missing non-standard ports list"
        shift
      ;;

      -d|--results-dir)
        RESULTS_DIR=$2
        shift
      ;;

      -s|--sids)
        SIDS_FILE=$2
        shift
      ;;

      *)
        first=$(echo $1 | cut -c1)
        [ "$first" = "-" ] && _exit "Unsupported script option: $1. Type '$SCR_BN -h' (or --help) to print the available options."
        PCAP_FILE=$1
      ;;
    esac
    shift
  done

  [ -z "$PCAP_FILE" ] && _exit "Missing pcap file"
  [ ! -f "$PCAP_FILE" ] && _exit "Cannot found provided pcap file '$PCAP_FILE' !!"
  [ -z "$RESULTS_DIR" ] && RESULTS_DIR=`dirname $PCAP_FILE`
  [ ! -d $RESULTS_DIR ] && _exit "The results directory '$RESULTS_DIR' must exists !!"
  if [ -n "$SIDS_FILE" ]
  then
    [ ! -f $SIDS_FILE ] && _exit "The Session-Id list file '$SIDS_FILE' provided, does not exist !!"
  fi
}

_exit () {
  echo
  echo -e $1
  echo

  # Cleanup
  rm -rf $tmpdir

  rc=1
  [ -n "$2" ] && rc=$2
  exit $rc
}


#############
# EXECUTION #
#############

echo
echo "============================================"
echo "Diameter buffer extractor from PCAP raw file"
echo "============================================"
echo

# Arguments:
[ "$1" = "" -o "$1" = "--help" -o "$1" = "-h" ] && usage
parse_arguments "$@"

# Tshark available:
which $TSHARK >/dev/null 2>/dev/null
[ $? -ne 0 ] && _exit "Missing 'tshark' tool !!"

# Get the frames with diameter content (take care about '-2' two-pass option and don't add it, because we need to get reassembled parts in their corresponding frames):
# Fields needed (we won't need diameter.hopbyhopid & diameter.endtoendid to verify diameter message as hint patterns; length management will be enough): see https://www.wireshark.org/docs/dfref/d/diameter.html
FIELDS="-e frame.number -e frame.time_epoch -e ip.src_host -e ip.dst_host -e diameter.cmd.code -e diameter.flags.request -e diameter.applicationId -e diameter.hopbyhopid -e diameter.endtoendid -e diameter.Session-Id -e diameter.Origin-Host -e diameter.Subscription-Id-Data -e diameter.Subscription-Id-Type -e frame.len -e tcp.len -e diameter.length -e frame.protocols -e tcp.segment -e sctp.chunk_length"
# Disect selectors for non-standard diameter ports:
for port in $OTHER_PORTS
do
  echo "Taking tcp port $port to be decoded as diameter protocol"
  DISECT_SELECTORS="$DISECT_SELECTORS -d tcp.port=$port,diameter"
done

$TSHARK -E separator="|" -r $PCAP_FILE -N mntC -Tfields $FIELDS $DISECT_SELECTORS 2>/dev/null > $tmpdir/all_frames
grep -i diameter $tmpdir/all_frames > $tmpdir/diameter_frames
#cat $tmpdir/all_frames

# Example output:
#                                                                                         / lengths \
# frm timestamp src   dst   code R  App-ID   HopByHop   EndToEnd  Sid OHost Subs SubsType FRM TCP DIA   protocol                              segments    sctp chunk length
#   1   2      3       4      5  6    7          8          9     10   11    12     13    14   15  16      17                                    18             19
#   1|tt.tt|gt_traf|vcbavipt|272|1|16777238|0x0004e6e6|0x000bd986|xxx|xxxxx|xxxxx|xxxxxxx|fff|432|432|eth:ip:tcp:diameter:diameter:diameter3gpp|
#   3|tt.tt|vcbavipt|gt_traf|272|0|16777238|0x0004e6e6|0x000bd986|xxx|xxxxx|xxxxx|xxxxxxx|fff|292|292|eth:ip:tcp:diameter:diameter:diameter3gpp|
#   5|tt.tt|gt_traf|vcbavipt||||||ffff|1400||eth:ip:tcp:diameter|
#   6|tt.tt|gt_traf|vcbavipt|265|1|16777236|0x000c73c3|0x0004cee4|xxx|xxxxx|xxxxx|xxxxxxx|fff|572|1972|eth:ip:tcp:diameter:diameter:diameter3gpp|5,6
#   8|tt.tt|vcbavipt|gt_traf|265|0|16777236|0x000c73c3|0x0004cee4|xxx|xxxxx|xxxxx|xxxxxxx|fff|248|248|eth:ip:tcp:diameter:diameter:diameter3gpp|
cat $tmpdir/diameter_frames | cut -d\| -f1 > $tmpdir/frames_num
all_frames=( $(cat $tmpdir/frames_num) )

# Segments groups (TCP or SCTP):
segmentsGroups=

# The case of SCTP is precalculated:
tshark -nr $PCAP_FILE $DISECT_SELECTORS -Tfields -e frame.number -e sctp.chunk_flags 2>/dev/null | awk '{ if ( $2 !~ "0x03" && $2 != "" ) print $1 }' > $tmpdir/extra_frames_num
if [ -s $tmpdir/extra_frames_num ]
then
  # Take into account these frames in all_frames array:

  cat $tmpdir/extra_frames_num >> $tmpdir/frames_num
  all_frames=( $(cat $tmpdir/frames_num | sort -un) )

  # Prepare as groups (1,2 7,8 ...)
  segmentsGroups=$(tshark -nr $PCAP_FILE $DISECT_SELECTORS -Tfields -e frame.number -e sctp.chunk_flags 2>/dev/null | awk '{ if ( $2 !~ "0x03" && $2 != "" ) print $1; else print "\n" }' | uniq | tr '\n' ',' | sed 's/,,/ /g')
fi


# Reassemble procedure (using frame 1 as example):
# (for non segmented frames, it is enough with tcp or diameter length within the frame content itself)
# 1) Get the TCP length: 432 bytes. 432*2 = 864 characters per byte in hexadecimal string format
# 2) Get the frame length: `wc -c $tmpdir/block.$frame` => 997
# 3) Get 864 from the tail: `cat $tmpdir/block.$frame | cut -c133

# Dump the hex blocks for all the diameter frames:
cat $PCAP_FILE | rawshark -s -r - -d proto:diameter -F data 2>/dev/null > $tmpdir/all_hex_data
for frame in ${all_frames[@]}; do
  grep "^$frame " $tmpdir/all_hex_data | cut -d\" -f2 | sed 's/://g' > $tmpdir/block.$frame
  frame_info=$(grep "^${frame}|" $tmpdir/diameter_frames)

  ##########################################################################################
  # Get the diameter part:
  frm_len=$(echo $frame_info | cut -d\| -f14)
  tcp_len=$(echo $frame_info | cut -d\| -f15)
  dia_len=$(echo $frame_info | cut -d\| -f16)
  protocol=$(echo $frame_info | cut -d\| -f17)

  # TCP and SCTP supported:
  transport=$(echo $protocol | grep -ow tcp)
  [ -z "$transport" ] && transport=$(echo $protocol | grep -ow sctp)
  # in case of segmented SCTP frames, protocol won't be sctp ...:
  if [ -s "$tmpdir/extra_frames_num" -a -z "$transport" ]
  then
    grep -qw ^${frame} $tmpdir/extra_frames_num
    if [ $? -eq 0 ]
    then
      transport=sctp
      frame_info=$(grep "^${frame}|" $tmpdir/all_frames)
    fi
  fi

  case $transport in
    sctp)
      chunk_length=$(echo $frame_info | cut -d\| -f19)
      dia_partial_len=$((chunk_length-16))
      if [ "$dia_len" != "$dia_partial_len" ]
      then
        #echo "Segmented SCTP frame $frame; taking partial diameter length = $dia_partial_len"
        dia_len=$dia_partial_len
      fi 
      cut_len=$((2*dia_len))
      cat $tmpdir/block.$frame | rev | cut -c-${cut_len} | rev > $RESULTS_DIR/$frame.hex
      ;;

    tcp)
      cut_len=$((2*(frm_len-tcp_len) + 1))
      cat $tmpdir/block.$frame | cut -c${cut_len}- > $RESULTS_DIR/$frame.hex
      cp $tmpdir/block.$frame /tmp/block
      segments=$(echo $frame_info | cut -d\| -f18)
      segmentsGroups="$segmentsGroups $segments"
      ;;

    *) _exit "Only TCP and SCTP transports supported !!"
      ;;
  esac
  ##########################################################################################

  echo -n "Created $RESULTS_DIR/$frame.hex"

  # Metadata:
  ts=$(echo $frame_info | cut -d\| -f2)
  date=$(date -d @$ts)
  src=$(echo $frame_info | cut -d\| -f3)
  dst=$(echo $frame_info | cut -d\| -f4)
  code=$(echo $frame_info | cut -d\| -f5)
  isreq=$(echo $frame_info | cut -d\| -f6)
  appid=$(echo $frame_info | cut -d\| -f7)
  sid=$(echo $frame_info | cut -d\| -f10)
  oh=$(echo $frame_info | cut -d\| -f11)
  subscriber=$(echo $frame_info | cut -d\| -f12)
  subscribertype=$(echo $frame_info | cut -d\| -f13)
  [ "$subscribertype" = "0" ] && subscribertype=msisdn
  [ "$subscribertype" = "1" ] && subscribertype=imsi


  #hbh=$(echo $frame_info | cut -d\| -f8)
  #e2e=$(echo $frame_info | cut -d\| -f9)
  # HBH and ETE To decimal:
  #hbh=$(printf "%d\n" $hbh)
  #e2e=$(printf "%d\n" $e2e)
  echo "date=$date" > $RESULTS_DIR/$frame.metadata
  echo "timestamp=$ts" >> $RESULTS_DIR/$frame.metadata
  echo "src=$src" >> $RESULTS_DIR/$frame.metadata
  echo "dst=$dst" >> $RESULTS_DIR/$frame.metadata
  echo "code=$code" >> $RESULTS_DIR/$frame.metadata
  echo "isrequest=$isreq" >> $RESULTS_DIR/$frame.metadata
  echo "applicationid=$appid" >> $RESULTS_DIR/$frame.metadata
  [ -n "$sid" ] && echo "sessionid=$sid" >> $RESULTS_DIR/$frame.metadata
  echo "originhost=$oh" >> $RESULTS_DIR/$frame.metadata
  if [ -n "$subscriber" ]
  then
    echo "subscriber=$subscriber" >> $RESULTS_DIR/$frame.metadata
    echo "subscribertype=$subscribertype" >> $RESULTS_DIR/$frame.metadata
  fi
  #echo "hopbyhop=$hbh" >> $RESULTS_DIR/$frame.metadata
  #echo "endtoend=$e2e" >> $RESULTS_DIR/$frame.metadata

  echo " and $RESULTS_DIR/$frame.metadata"
done

# Join frames which need to be reassembled:
to_delete=
for group in $segmentsGroups
do
  group_array=( $(echo $group | sed 's/,/ /g') )
  echo "Grouping frames ${group_array[*]} ..."
  for frame in ${group_array[@]}; do
    cat $RESULTS_DIR/$frame.hex >> $tmpdir/diam.$group 2>/dev/null
  done
  cat $tmpdir/diam.$group | tr -d '\n' > $RESULTS_DIR/$frame.hex
  # Delete all frames except last one in the group:
  to_delete="$to_delete  $(echo ${group_array[*]} | awk '{$NF=""; print $0}')"
done

# Delete superfluous metadata:
#segments=( $(cat $tmpdir/diameter_frames | awk -F\| '{ if ($16 == "") print $1 }') )
segments=( $(echo $to_delete) )
echo "Deleting superfluous buffers & metadata (${segments[*]}) ..."
for s in ${segments[@]}; do rm -f $RESULTS_DIR/${s}.* ; done

# Basic checking:
ls $RESULTS_DIR/*.metadata >/dev/null 2>/dev/null
[ $? -ne 0 ] && { echo "No metadata generated !" ; exit 1 ; }

# Detecting Session-Id values:
grep ^sessionid= $RESULTS_DIR/*.metadata 2>/dev/null | cut -d= -f2- | sort -u > $RESULTS_DIR/session-ids
if [ -s $RESULTS_DIR/session-ids ]
then
  count=0
  while read -r line; do count=$((count+1)) ; echo "Detected Session-Id $count:  $line"; done < $RESULTS_DIR/session-ids
  rm $RESULTS_DIR/session-ids
fi

# Detecting Origin-Host values:
grep ^originhost= $RESULTS_DIR/*.metadata 2>/dev/null | cut -d= -f2- | sort -u > $RESULTS_DIR/origin-hosts
if [ -s $RESULTS_DIR/origin-hosts ]
then
  count=0
  while read -r line; do count=$((count+1)) ; echo "Detected Origin-Host $count: $line"; done < $RESULTS_DIR/origin-hosts
  #rm $RESULTS_DIR/origin-hosts
fi

# Purge frames with Session-Id not wanted:
if [ -n "$SIDS_FILE" ]
then
  grep -l -w -f $SIDS_FILE $RESULTS_DIR/*metadata > $RESULTS_DIR/.wanted
  grep -l ^sessionid $RESULTS_DIR/*metadata > $RESULTS_DIR/.all
  for file in `grep -vf $RESULTS_DIR/.wanted $RESULTS_DIR/.all`
  do
    frm=$(basename $file | cut -d\. -f1)
    sid=$(grep ^sessionid= $file | cut -d= -f2-)
    echo "Purge results for frame $frm (Session-Id: '$sid') ..."
    rm $RESULTS_DIR/${frm}.*
  done
  rm $RESULTS_DIR/.wanted $RESULTS_DIR/.all
fi

# Purge frames with invalid metadata:
invalid=( $(grep -lw "^code=$" $RESULTS_DIR/*metadata) )
for file in ${invalid[@]}
do
  frm=$(basename $file | cut -d\. -f1)
  rm $RESULTS_DIR/${frm}.*
done

_exit "Done!" 0

