// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


#ifndef anna_xml_Document_hpp
#define anna_xml_Document_hpp

struct _xmlDoc;
struct _xmlValidCtxt;
struct _xmlDoc;

#include <anna/core/DataBlock.hpp>
#include <anna/core/RuntimeException.hpp>

namespace anna {

namespace xml {

class Parser;
class XPath;
class DTD;
class Node;

/**
   Clase base para manejar los documentos XML.

   Ejemplo de documento XML:

   \code

   <xvc HeartBeat="20000">
      <broadcast>
         <INetAddress Address="204.152.65.15" Port="2000"/>
         <INetAddress Address="204.152.65.47" Port="2002"/>
      </broadcast>

      <ethernet Mode="raw" VirtualAddress="1.1.1.100">
         <Input Device="/dev/qfe" PhysicalAccessPoint="2" MAC="8:0:20:9e:ee:21"/>
         <Output Device="/dev/qfe" PhysicalAccessPoint="2" MAC="8:0:20:9e:ee:c9"/>
      </ethernet>
   </xvc>

   \endcode
*/
class Document : anna::DataBlock {
public:
  /**
     Destructor.
  */
  virtual ~Document();

  /**
     Inicializa el contenido del documento XML.
     \param content Contenido de la Documento, depedendiendo del tipo de Documento hara referencia
     a un nombre de archivo, una URI o a una cadena.
  */
  void initialize(const char* content) throw(RuntimeException);

  /**
     Inicializa el contenido del documento XML.
     \param content Contenido del Documento XML.
  */
  void initialize(const DataBlock& content) throw(RuntimeException);

  /**
   * Obtiene el conjunto de caracteres usado en el documento XML. Puede ser NULL.
   * \return El conjunto de caracteres usado en el documento XML.
   */
  const char* getEncoding() const throw();

  /**
   * Obtiene la versin indicada en el documento XML.Puede ser NULL.
   * \return La versin indicada en el documento XML.
   */
  const char* getVersion() const throw();

  /**
     Devuelve el contenido asociado al documento XML.
     \return El contenido asociado al documento XML.
  */
  virtual const DataBlock& getContent() const throw(RuntimeException) { return *this; }

  /**
     Devuelve el contenido asociado al documento XML expresado como una cadena C.
     \return El contenido asociado al documento XML expresado como una cadena C.
  */
  const char* getContentAsCString() const throw(RuntimeException);

  /**
   * Establece el conjunto de caracteres usado en el documento XML.
   * \param encoding Literal que indica el conjunto de caracteres. Puede ser NULL.
   */
  void setEncoding(const char* encoding) throw();

  /**
   * Establece la versin usada en el documento XML.
   * \param encoding Literal que indica la versin. Puede ser NULL.
   */
  void setVersion(const char* version) throw();

  /**
   * Analiza el contenido del documento XML y devuelve el nodo raz.
   * \return El nodo raiz del arbol XML correspondiente al resultado del analisis.
   * \warning Este documento debe estar correctamente inicializado.
   */
  const xml::Node* parse() throw(RuntimeException);

  /**
   * Analiza el documento XML recibido como parametro, y verifica que cumpla las reglas
   * establecidas por la DTD.
   * \param dtd DTD que debe cumplir el documento XML.
   * \return El nodo raiz del arbol XML correspondiente al resultado del analisis.
   * \warning Este documento debe estar correctamente inicializado.
   */
  const xml::Node* parse(const DTD& dtd) throw(RuntimeException);

protected:
  /**
     Constructor.
  */
  Document();

  /**
   * Libera la memoria asociada a los componentes de este documento.
   */
  void clear() throw();

  /*
   * Establece el contenido de este documento XML.
   * \param content Buffer que contiene una C-String con la que inciar este documento XML.
   */
  void setContent(const char* content) throw() {
    DataBlock::clear();
    DataBlock::append(content, anna_strlen(content) + 1);
    a_contentIsCString = true;
  }

  /*
   * Establece el contenido de este documento XML.
   * \param content Buffer que apunta al contenido con la que inciar este documento XML.
   * \param size Longitud del buffer.
   * \warning Slo uso interno
   */
  void setContent(const char* content, const int size) throw() {
    DataBlock::clear();
    DataBlock::append(content, size);
    a_contentIsCString = false;
  }

  /*
   * Establece el contenido de este documento XML.
   * \param content Bloque de datos del que copiar el valor para iniciar este documento XML.
   */
  void setContent(const DataBlock& content) throw() {
    DataBlock::operator= (content);
    a_contentIsCString = false;
  }

private:
  _xmlDoc* a_handle;
  std::string* a_encoding;
  std::string* a_version;
  bool a_contentIsCString;
  DataBlock* a_asCString;
  Parser* a_parser;

  Document(const Document&);

  virtual _xmlDoc* do_initialize(const char* content) throw(RuntimeException) = 0;
  virtual _xmlDoc* do_initialize(const DataBlock& content) throw(RuntimeException) = 0;

  friend class Parser;
  friend class XPath;
};

}
}

#endif
