// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


#include <anna/core/tracing/Logger.hpp>
#include <anna/core/tracing/TraceMethod.hpp>

#include <anna/comm/Communicator.hpp>
#include <anna/comm/ClientSocket.hpp>
#include <anna/comm/Transport.hpp>
#include <anna/comm/Device.hpp>
#include <anna/comm/Receiver.hpp>

#include <anna/comm/handler/MetaClientSocket.hpp>

using namespace std;
using namespace anna;

/*
 * Este mtodo se ejecuta desde un nico thread (Tx). Cada socket tiene asociado un thread que lo trata.
 */
void comm::handler::MetaClientSocket::apply()
throw(RuntimeException) {
  LOGMETHOD(TraceMethod tm(Logger::Local7, "comm::handler::MetaClientSocket", "apply", ANNA_FILE_LOCATION));
  comm::ClientSocket* clientSocket = getClientSocket();

  if(clientSocket == NULL) {
    string msg("comm::handler::MetaClientSocket::apply | fd: ");
    msg += functions::asString(getfd());
    msg += " | getClientSocket returns NULL";
    Logger::warning(msg, ANNA_FILE_LOCATION);
    a_communicator->detach(this);
    return;
  }

  /*
   * En MT se ha comprobado que hay alguna combinaci�n en la que se llega a tratar un socket que ha sido cerrado previamente.
   */
  if(clientSocket->isOpened() == false) {
    string msg("comm::handler::MetaClientSocket::apply | fd: ");
    msg += functions::asString(getfd());
    msg += " | Detected incoming socket overload";
    Logger::error(msg, ANNA_FILE_LOCATION);
    requestStop();
    return;
  }

  bool mustDetach(false);
  {
    Guard guard(clientSocket, "comm::ClientSocket from comm::handler::MetaClientSocket::apply");
    comm::ClientSocket::Notify::_v notify = clientSocket->receive();

    if(notify == comm::ClientSocket::Notify::ReceiveData) {
      const Message* message;
      const DataBlock* dataBlock;
      int messageCounter = 0;
      Transport* transport = clientSocket->getTransport();
      Receiver* receiver = clientSocket->getReceiver();
      clientSocket->activate(comm::ClientSocket::Status::Working);

      while(canContinue() == true && (dataBlock = clientSocket->fetch()) != NULL) {
        try {
          try {
            message = transport->decode(*dataBlock);                          // (4)
          } catch(RuntimeException& ex) {
            ex.trace();
            clientSocket->activate(comm::ClientSocket::Status::Corrupt);
            message = NULL;
            break;
          }

          if(message != NULL) {
            messageCounter ++;

            if(receiver == NULL)
              a_communicator->eventReceiveMessage(*clientSocket, *message);
            else
              receiver->apply(*clientSocket, *message);
          }
        } catch(RuntimeException& ex) {
          ex.trace();
        }
      }

      LOGLOCAL6(
        string msg("comm::handler::MetaClientSocket::apply | fd: ");
        msg += functions::asString(getfd());
        msg += functions::asText(" | Message Counter: ", messageCounter);
        Logger::write(Logger::Local6, msg, ANNA_FILE_LOCATION);
      );
      clientSocket->deactivate(comm::ClientSocket::Status::Working);

      if(clientSocket->hasRequestedClose() == true || clientSocket->isCorrupt()) {
        if(clientSocket->isCorrupt()) {
          string msg(clientSocket->asString());
          msg += " | ClientSocket was closed because of wrong message";
          Logger::error(msg, ANNA_FILE_LOCATION);
          a_communicator->eventDiscardConnection(*clientSocket);
        }

        mustDetach = true;
      }
    } else if(notify == comm::ClientSocket::Notify::Close || notify == comm::ClientSocket::Notify::Corrupt) {  // (2)
      if(clientSocket->isCorrupt()) {
        string msg(clientSocket->asString());
        msg += " | ClientSocket was closed because of wrong message";
        Logger::error(msg, ANNA_FILE_LOCATION);
        a_communicator->eventDiscardConnection(*clientSocket);
      }

      mustDetach = true;
    }
  }

  /*
   * Termina llamando al XXX::finalize sobre-escrito por la clase heredada de esta
   */
  if(mustDetach == true)
    a_communicator->detach(this);
}

//virtual
bool comm::handler::MetaClientSocket::testClose()
throw(RuntimeException) {
  LOGMETHOD(TraceMethod tm(Logger::Local7, "comm::handler::MetaClientSocket", "testClose", ANNA_FILE_LOCATION));
  comm::ClientSocket* clientSocket = getClientSocket();

  if(clientSocket == NULL) {
    string msg("comm::handler::MetaClientSocket::testClose | fd: ");
    msg += functions::asString(getfd());
    msg += " | getClientSocket returns NULL";
    Logger::warning(msg, ANNA_FILE_LOCATION);
    a_communicator->detach(this);
    return false;
  }

  bool mustClose;
  {
    Guard guard(clientSocket, "comm::ClientSocket from comm::handler::MetaClientSocket::testClose");
    mustClose = clientSocket->hasRequestedClose();
  }

  /*
   * Termina llamando al XXX::finalize sobre-escrito por la clase heredada de esta
   */
  if(mustClose == true)
    a_communicator->detach(this);

  return mustClose;
}

/*
 * Este metodo se invoca cuando algun elemento externo detecta la cada de toda la red 'address'
 *
 * Por lo que se puede invocar desde cualquier otro thread no reconocido.
 */
void comm::handler::MetaClientSocket::breakAddress(const in_addr_t& address)
throw() {
  comm::ClientSocket* clientSocket = getClientSocket();

  if(clientSocket == NULL) {
    string msg(asString());
    msg += " | getClientSocket returns NULL";
    throw RuntimeException(msg, ANNA_FILE_LOCATION);
  }

  {
    Guard guard(clientSocket, "comm::ClientSocket from comm::handler::MetaClientSocket::breakAddress");
    const comm::AccessPoint& accessPoint = clientSocket->getRemoteAccessPoint();
    const comm::Device* device = accessPoint.getINetAddress().getDevice(false);

    if(device == NULL)
      return;

    if(device->getAddress() != address)
      return;

    LOGDEBUG(
      string msg("comm::handler::MetaClientSocket::breakAddress | ");
      msg += clientSocket->asString();
      Logger::debug(msg, ANNA_FILE_LOCATION);
    );
  }

  /*
   * Termina llamando al XXX::finalize sobre-escrito por la clase heredada de esta
   */
  a_communicator->detach(clientSocket);
}

