// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


#include <sys/time.h>
#include <fcntl.h>
#include <strings.h>
#include <sys/types.h>
#include <poll.h>

#ifdef __CYGWIN__
#include <sys/uio.h>
#endif

#include <anna/core/functions.hpp>
#include <anna/core/tracing/Logger.hpp>

#include <anna/comm/internal/BinderSocket.hpp>
#include <anna/comm/ServerSocket.hpp>

using namespace std;
using namespace anna;

//-------------------------------------------------------------------------------------------------
// No sabemos si tenemos que actuar como Cliente o Servidor; por lo que tenemos que establecer
// ambos extremos del Socket.
//-------------------------------------------------------------------------------------------------
comm::BinderSocket::BinderSocket(comm::ServerSocket* serverSocket) :
  ClientSocket(
    string("/tmp/").append(serverSocket->getLocalAccessPoint().serialize()),
    Type::Stream
  ),
  a_serverSocket(*serverSocket) {
}

//-------------------------------------------------------------------------------------------------
// (1) Si consigue hacer el bind => es el primero. A partir de ahora dara acceso al resto.

// (0) a_bindDelay esta expresando en milisegundos => maxDelay estara en nanosegundos.
// (0.1) Guardara los fds asociados al socket (real) y al socket por el que enviamos la informacion
//     para poder hacer multiples binds sobre una misma direccion.
// (0.2) Vamos a enviar 2 fds (el socket real y el socket UNIX).
// (1) El Communicator tendra que tener en cuenta que se trata de un socket bind-compartido
//     para meter el fd del publicador en la lista de descriptores por los que es posible que nos
//     lleguen mensajes.
// (2) Si falla el bind => cierra los recursos. Ya que nos llegara el nuevo fd atraves del mensaje
//     del recvmsg.
// (3) Conectamos como cliente con el socket UNIX => cuando la parte servidora (a traves del
//     Communicator) detecte nuestra presencia deberia enviarnos la informacion para que podamos
//     crear el socket con el bind-compartido. Limpiamos la direccion local para evitar hacer otro
//      bind
// (4) Si el 'connect' falla => se cierra el socket automaticamente => lo abre de nuevo.
// (5) "Cierra" el socket UNIX ya que no volvera a usarlo. A partir de ahora el valor bueno sera
//     el indicado en la estructura que nos pasa alguno de los servidores que estan escuchando
//     el socket UNIX.
//
// Visto en: http://mail-index.netbsd.org/current-users/1998/01/16/0010.html
//-------------------------------------------------------------------------------------------------
void comm::BinderSocket::requestBind(const struct sockaddr* s, const  int len)
throw(RuntimeException) {
  if(::bind(a_serverSocket.a_fd, s, len) != -1) {                // (1)
    unlink(getRemoteAccessPoint().getPath().c_str());
    this->bind();
    anna_socket_assert(::listen(a_fd, 10) == -1, "Error preparing listen access point");
    return;
  }

  const int xerrno = errno;

  a_serverSocket.close();                                        // (2)

  if(xerrno != EADDRINUSE)
    throw RuntimeException(asString(), xerrno, ANNA_FILE_LOCATION);

  this->a_localAccessPoint.clear();                           // (3)
  this->connect();                                               // (4)
  waitBind(a_serverSocket.getBindDelay());
}

//-------------------------------------------------------------------------------------------------
// Este metodo se invoca desde el Communicator cuando se detecta activada en el socket UNIX
// empleado para comunicar todos los procesos que quieran compartir una determinada direccion.
//
// (1) Acepta la conexion por el socket UNIX. Ver (3) de sharedBind.
// (2) En el IOV debemos mandar al menos 1 byte.
// (3) Rellenamos la estructura que recojera algun otro proceso que esta esperando a leer del
//     socket UNIX.
// (4) Cierra la nueva conexion ya que no vamos a volver a usarla para nada.
//-------------------------------------------------------------------------------------------------
void comm::BinderSocket::responseBind()
throw(RuntimeException) {
  struct iovec iov;
  struct msghdr mh;
  int fds [2];
  int garbage(0);
  sockaddr_un s;
  socklen_t len(sizeof(sockaddr_un));
  anna_memset(&s, 0, sizeof(sockaddr_un));
  int newSocket = ::accept(a_fd, (sockaddr*) & s, &len);      // (1)

  if(newSocket == -1) {
    const int xerrno = errno;
    throw RuntimeException(asString(), xerrno, ANNA_FILE_LOCATION);
  }

  iov.iov_len = sizeof(garbage);                                                   // (2)
  iov.iov_base = (char*) & garbage;
  mh.msg_iov = &iov;
  mh.msg_iovlen = 1;
  mh.msg_name = NULL;
  mh.msg_namelen = 0;
  fds [0] = a_serverSocket.a_fd;                                                  // (3)
  fds [1] = a_fd;
#ifndef __sun
  struct cmsghdr cmh;
  char buffer [sizeof(struct cmsghdr) + sizeof(fds)];
  mh.msg_control = &buffer;
  mh.msg_controllen = cmh.cmsg_len = sizeof(struct cmsghdr) + sizeof(fds);
  cmh.cmsg_level = SOL_SOCKET;
  cmh.cmsg_type = SCM_RIGHTS;
  anna_memcpy(buffer, &cmh, sizeof(cmh));
  anna_memcpy(buffer + sizeof(struct cmsghdr), fds, sizeof(fds));
  mh.msg_flags = 0;
#else
  mh.msg_accrights = (caddr_t) fds;
  mh.msg_accrightslen = sizeof(fds);
#endif

  if(sendmsg(newSocket, &mh, 0) < 0) {
    int xerrno = errno;
    ::close(newSocket);
    throw RuntimeException(asString(), xerrno, ANNA_FILE_LOCATION);
  }

  ::close(newSocket);                                                              // (4)
  LOGDEBUG(
    string msg("anna::comm::BinderSocket::responseBind | ");
    msg += functions::asText("ServerSocket's fd: ", fds [0]);
    msg += functions::asText(" | BinderSocket's fd: ", fds [1]);
    Logger::debug(msg, ANNA_FILE_LOCATION);
  );
}

void comm::BinderSocket::waitBind(const Millisecond &maxDelay)
throw(RuntimeException) {
  struct msghdr mh;
  struct iovec iov;
  const int garbage(0);
  int fds [2];
  iov.iov_len = sizeof(garbage);
  iov.iov_base = (char*) & garbage;
  mh.msg_iov = &iov;
  mh.msg_iovlen = 1;
  mh.msg_name = NULL;
  mh.msg_namelen = 0;
#ifndef __sun
  char buffer [sizeof(struct cmsghdr) + sizeof(fds)];
  mh.msg_control = &buffer;
  mh.msg_controllen = sizeof(struct cmsghdr) + sizeof(fds);
  mh.msg_flags = 0;
#else
  mh.msg_accrights = (caddr_t) fds;
  mh.msg_accrightslen = sizeof(fds);
#endif
  pollfd poll;
  int r;
  poll.fd = a_fd;
  poll.events = POLLIN | POLLRDNORM;

  if(::poll(&poll, 1, maxDelay) != 1) {
    string msg(a_serverSocket.asString());
    msg += " | No data is received from shared bind";
    throw RuntimeException(msg, ANNA_FILE_LOCATION);
  }

  r = recvmsg(a_fd, &mh, 0);

  if(r <= 0) {
    const int xerrno = errno;
    std::string msg(a_serverSocket.asString());
    msg += " | ";
    msg += asString();

    if(r == -1) {
      msg += " | Error in shared bind reception";
      throw RuntimeException(msg, xerrno, ANNA_FILE_LOCATION);
    } else if(r == 0) {
      msg += " | No item received to obtain shared bind";
      throw RuntimeException(msg, ANNA_FILE_LOCATION);
    }
  }

  close();
#ifndef __sun
  anna_memcpy(&fds, buffer + sizeof(cmsghdr), sizeof(fds));
#endif
  a_serverSocket.a_fd = fds [0];
  a_fd = fds [1];
  LOGDEBUG(
    string msg("anna::comm::BinderSocket::waitBind | ");
    msg += functions::asText("ServerSocket's fd: ", fds [0]);
    msg += functions::asText(" | BinderSocket's fd: ", fds [1]);
    Logger::debug(msg, ANNA_FILE_LOCATION);
  );
}

