// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


// Local
#include <anna/diameter/codec/Avp.hpp>
#include <anna/diameter/codec/Format.hpp>

#include <anna/config/defines.hpp> // general types, decoding helpers (DECODE[2/3/4]BYTES_INDX_VALUETYPE), etc.
#include <anna/diameter/functions.hpp>
#include <anna/diameter/helpers/defines.hpp>
#include <anna/diameter/codec/basetypes/basetypes.hpp>
#include <anna/diameter/codec/functions.hpp> // REQUIRED_WORDS
#include <anna/diameter/codec/OamModule.hpp>
#include <anna/diameter/stack/Avp.hpp>
#include <anna/diameter/stack/Format.hpp>
#include <anna/diameter/stack/Dictionary.hpp>
#include <anna/diameter/stack/Engine.hpp>
#include <anna/diameter/codec/Engine.hpp>
#include <anna/core/functions.hpp>
#include <anna/core/util/RegularExpression.hpp>

#include <anna/core/tracing/Logger.hpp>
#include <anna/core/functions.hpp>
#include <anna/core/tracing/TraceMethod.hpp>
#include <anna/xml/xml.hpp>

// STL
#include <map>


using namespace anna;
using namespace anna::diameter::codec;


//static
namespace anna {

namespace diameter {

namespace codec {
const int Avp::HeaderLengthVactive(12);
const int Avp::HeaderLengthVinactive(8);
const U8 Avp::VBitMask(0x80);
const U8 Avp::MBitMask(0x40);
const U8 Avp::PBitMask(0x20);
}
}
}

//------------------------------------------------------------------------------
//------------------------------------------------------------------- Avp::Avp()
//------------------------------------------------------------------------------
Avp::Avp(Engine *engine) : a_engine(engine) {
  initialize();
}

//------------------------------------------------------------------------------
//------------------------------------------------------------------- Avp::Avp()
//------------------------------------------------------------------------------
Avp::Avp(AvpId id, Engine *engine) : a_engine(engine) {
  initialize();
  setId(id);
}

//------------------------------------------------------------------------------
//------------------------------------------------------------------ Avp::~Avp()
//------------------------------------------------------------------------------
Avp::~Avp() {
  clear();
}

//------------------------------------------------------------------------------
//------------------------------------------------------------- Avp::setEngine()
//------------------------------------------------------------------------------
void Avp::setEngine(Engine *engine) throw() {
  if (a_engine && engine != a_engine) {
    LOGWARNING(anna::Logger::warning("Ignored: it is not a good practice to change the codec engine once assigned. Clear the avp first to set the engine again.", ANNA_FILE_LOCATION));
    return;
  }

  a_engine = engine;
}

//------------------------------------------------------------------------------
//------------------------------------------------------------- Avp::getEngine()
//------------------------------------------------------------------------------
Engine * Avp::getEngine() const throw(anna::RuntimeException) {
  if(!a_engine)
    throw anna::RuntimeException("Invalid codec engine reference (NULL). Use setEngine() to set the corresponding codec engine", ANNA_FILE_LOCATION);

  return a_engine;
}


//------------------------------------------------------------------------------
//------------------------------------------------------------ Avp::initialize()
//------------------------------------------------------------------------------
void Avp::initialize() throw() {
  a_id = helpers::AVPID__AVP; // (0,0)
  a_flags = 0x00;
  a_insertionPositionForChilds = 0;
  a_OctetString = NULL;
  a_Integer32 = NULL;
  a_Integer64 = NULL;
  a_Unsigned32 = NULL;
  a_Unsigned64 = NULL;
  a_Float32 = NULL;
  a_Float64 = NULL;
  //a_avps.clear();
  a_Address = NULL;
  a_Time = NULL;
  a_UTF8String = NULL;
  a_DiameterIdentity = NULL;
  a_DiameterURI = NULL;
  a_Enumerated = NULL;
  a_IPFilterRule = NULL;
  a_QoSFilterRule = NULL;
  a_Unknown = NULL;
  //a_finds.clear();
  /////////////////////
  // Format specific //
  /////////////////////
  initializeByFormat();
}


//------------------------------------------------------------------------------
//----------------------------------------------------------------- Avp::clear()
//------------------------------------------------------------------------------
void Avp::clear() throw(anna::RuntimeException) {
  delete a_OctetString;
  delete a_Integer32;
  delete a_Integer64;
  delete a_Unsigned32;
  delete a_Unsigned64;
  delete a_Float32;
  delete a_Float64;

  for(avp_iterator it = avp_begin(); it != avp_end(); it++) { /*avp(it)->clear(); */getEngine()->releaseAvp(avp(it)); }

  a_avps.clear();
  delete a_Address;
  delete a_Time;
  delete a_UTF8String;
  delete a_DiameterIdentity;
  delete a_DiameterURI;
  delete a_Enumerated;
  delete a_IPFilterRule;
  delete a_QoSFilterRule;
  delete a_Unknown;
  // Cache system:
  a_finds.clear();
  /////////////////////
  // Format specific //
  /////////////////////
  clearByFormat();
  // Initialize:
  initialize();
}


//------------------------------------------------------------------------------
//-------------------------------------------------------------- Avp::avp_find()
//------------------------------------------------------------------------------
avp_iterator Avp::avp_find(avp_container &avps, AvpId id, unsigned int position) throw() {
  int match = 0;
  Avp *aux;

  for(avp_iterator it = avps.begin(); it != avps.end(); it++) {
    aux = (*it).second;

    if(aux && (aux->getId() == id)) {
      match++;

      if(match == position) return it;
    }
  }

  return avps.end();
}


//------------------------------------------------------------------------------
//---------------------------------------------------------------- Avp::addAvp()
//------------------------------------------------------------------------------
Avp * Avp::addAvp(avp_container &avps, int &insertionPositionForChilds, AvpId id, Engine *engine) throw() {
  Avp * result = engine->createAvp(NULL);
  result->setId(id);
  addChild(avps, insertionPositionForChilds, result);
  return result;
}


//------------------------------------------------------------------------------
//---------------------------------------------------------------- Avp::addAvp()
//------------------------------------------------------------------------------
Avp * Avp::addAvp(Avp * avp) throw(anna::RuntimeException) {
  if(!avp) return NULL;
  if (avp->getEngine() != getEngine()) return NULL;
  addChild(avp);
  return avp;
}


//------------------------------------------------------------------------------
//------------------------------------------------------------- Avp::removeAvp()
//------------------------------------------------------------------------------
bool Avp::removeAvp(avp_container &avps, find_container &finds, AvpId id, int ocurrence, Engine *engine) throw() {
  bool removed = false;
  bool do_remove = true;
  int position = ocurrence;

  if(position < 0) {  /* reversed access */
    position = countAvp(avps, id) + 1 + position;

    // Special cases -> 0 and negative results:
    // (1) User wants to remove the last, but no avp is found: position = 0 + 1 - 1 = 0 (would removes all but, no avp will be found: return now)
    // (2) User wants to remove some intermediate avp, i.e. '-5', but only 4 avps are found: 4 + 1 - 5 = 0 (would removes all !!!: return to avoid unexpected behaviour)
    // (2) User wants to remove some intermediate avp, i.e. '-5', but only 3 avps are found: 3 + 1 - 5 = -1 (can't find negative positions ...)
    if(position <= 0) do_remove = false;
  }

  if(do_remove) {
    int n = position ? position : 1 /* will search always the first and remove, the first and remove, the first and remove ... */;
    avp_iterator it;

    while((it = avp_find(avps, id, n)) != avps.end()) {
      engine->releaseAvp((*it).second);
      avps.erase(it);
      removed = true;

      if(position) break;
    }
  }

  if(removed) {
    // Cache system reset (remove will invalidate cached findings):
    finds.clear();
    LOGDEBUG(
      const stack::Avp *stackAvp = getStackAvp(id, engine);
      std::string msg = "Removed Avp "; msg += stackAvp ? (stackAvp->getName()) : (anna::diameter::functions::avpIdAsPairString(id));
      msg += anna::functions::asString(" with provided ocurrence (%d)", ocurrence);
      anna::Logger::debug(msg, ANNA_FILE_LOCATION);
    );
  } else {
    LOGDEBUG(
      const stack::Avp *stackAvp = getStackAvp(id, engine);
      std::string msg = "Can't found Avp "; msg += stackAvp ? (stackAvp->getName()) : (anna::diameter::functions::avpIdAsPairString(id));
      msg += anna::functions::asString(" with provided ocurrence (%d), in order to be removed", ocurrence);
      anna::Logger::debug(msg, ANNA_FILE_LOCATION);
    );
  }

  return removed;
}


//------------------------------------------------------------------------------
//-------------------------------------------------------------- Avp::countAvp()
//------------------------------------------------------------------------------
int Avp::countAvp(const avp_container &avps, AvpId id) throw() {
  int result = 0;
  const_avp_iterator it;

  while((it = avp_find(avps, id, result + 1)) != avps.end()) result++;

  return result;
}


//------------------------------------------------------------------------------
//-------------------------------------------------------------- Avp::firstAvp()
//------------------------------------------------------------------------------
const Avp* Avp::firstAvp(const avp_container &avps, AvpId id) throw() {
  const_avp_iterator it = avp_find(avps, id, 1);

  if(it != avps.end())
    return (*it).second;

  return NULL;
}


//------------------------------------------------------------------------------
//----------------------------------------------------------- Avp::countChilds()
//------------------------------------------------------------------------------
int Avp::countChilds(const avp_container &avps) throw() {
  return avps.size();
}


//------------------------------------------------------------------------------
//---------------------------------------------------------------- Avp::getAvp()
//------------------------------------------------------------------------------
const Avp *Avp::getAvp(const avp_container &avps, find_container &finds, AvpId id, int ocurrence, Engine *engine, anna::Exception::Mode::_v emode) throw(anna::RuntimeException) {
  if(ocurrence == 0) {
    LOGDEBUG(anna::Logger::debug("Ocurrence number zero has no sense. NULL returned.", ANNA_FILE_LOCATION));
    return NULL;
  }

  bool do_search = true;
  bool cached;
  int position = ocurrence;

  if(position < 0) {  /* reversed access */
    position = countAvp(avps, id) + 1 + position;

    // Special cases -> 0 and negative results:
    // (1) User wants to get the last, but no avp is found: position = 0 + 1 - 1 = 0 (can't find 0-position ...)
    // (2) User wants to get some intermediate avp, i.e. '-5', but only 4 avps are found: 4 + 1 - 5 = 0 (can't find 0-position ...)
    // (2) User wants to get some intermediate avp, i.e. '-5', but only 3 avps are found: 3 + 1 - 5 = -1 (can't find negative positions ...)
    if(position <= 0) do_search = false;
  }

  const Avp * result = NULL;

  if(do_search) {
    // Search at cache finds:
    find_key key(id, position);
    find_iterator fit = finds.find(key);
    cached = (fit != finds.end());

    if(cached) {
      result = (*fit).second;
    } else {
      const_avp_iterator it = avp_find(avps, id, position);

      if(it != avps.end()) {
        result = (*it).second;
        finds[key] = const_cast<Avp*>(result); // store in cache
      }
    }
  }

  // Avp found:
  if(result) {
    LOGDEBUG(
      const stack::Avp *stackAvp = getStackAvp(id, engine);
      std::string msg = "Found Avp "; msg += stackAvp ? (stackAvp->getName()) : (anna::diameter::functions::avpIdAsPairString(id));
      msg += anna::functions::asString(" with provided ocurrence (%d) which was ", ocurrence);
      msg += cached ? "already cached:\n" : "the first search, now cached:\n";
      msg += result->asXMLString();
      anna::Logger::debug(msg, ANNA_FILE_LOCATION);
    );
    return result;
  }

  // Avp not found:
  switch(emode) {
  case anna::Exception::Mode::Throw: {
    const stack::Avp *stackAvp = getStackAvp(id, engine);
    std::string msg = "Can't found Avp "; msg += stackAvp ? (stackAvp->getName()) : (anna::diameter::functions::avpIdAsPairString(id));
    msg += anna::functions::asString(" with provided ocurrence (%d)", ocurrence);
    throw anna::RuntimeException(msg, ANNA_FILE_LOCATION);
    break;
  }
  case anna::Exception::Mode::Trace: {
    LOGWARNING(
      const stack::Avp *stackAvp = getStackAvp(id, engine);
      std::string msg = "Can't found Avp "; msg += stackAvp ? (stackAvp->getName()) : (anna::diameter::functions::avpIdAsPairString(id));
      msg += anna::functions::asString(" with provided ocurrence (%d)", ocurrence);
      anna::Logger::warning(msg, ANNA_FILE_LOCATION);
    );
    break;
  }
  case anna::Exception::Mode::Ignore: {
    LOGDEBUG(
      const stack::Avp *stackAvp = getStackAvp(id, engine);
      std::string msg = "Can't found Avp "; msg += stackAvp ? (stackAvp->getName()) : (anna::diameter::functions::avpIdAsPairString(id));
      msg += anna::functions::asString(" with provided ocurrence (%d)", ocurrence);
      anna::Logger::debug(msg, ANNA_FILE_LOCATION);
    );
    break;
  }
  }

  return NULL;
}


//------------------------------------------------------------------------------
//--------------------------------------------------------------- Avp::_getAvp()
//------------------------------------------------------------------------------
const Avp *Avp::_getAvp(AvpId id, int ocurrence, anna::Exception::Mode::_v emode) const throw(anna::RuntimeException) {
  // Dictionary stack avp and format (this):
  const stack::Avp *stackAvp = getStackAvp();
  const stack::Format *stackFormat = stackAvp ? (stackAvp->getFormat()) : NULL /*Unknown*/;

  if(!stackFormat || !stackFormat->isGrouped())
    throw anna::RuntimeException("This method only applies over grouped avps.", ANNA_FILE_LOCATION);

  return getAvp(a_avps, (find_container&)a_finds, id, ocurrence, getEngine(), emode);
}


//------------------------------------------------------------------------------
//---------------------------------------------------------- Avp::assertFormat()
//------------------------------------------------------------------------------
void Avp::assertFormat(const std::string &name) const throw(anna::RuntimeException) {
  // Dictionary stack avp and format:
  const stack::Avp *stackAvp = getStackAvp();
  const stack::Format *stackFormat = stackAvp ? (stackAvp->getFormat()) : NULL /*Unknown*/;
  std::string format = stackFormat ? stackFormat->getName() : "Unknown";

  if(format != name) {
    std::string msg = "The Avp ";
    msg += anna::diameter::functions::avpIdAsPairString(a_id);
    msg += " with format '";
    msg += format;
    msg += "' is not compatible with the API method used (";
    msg += name;
    msg += ")";
    throw anna::RuntimeException(msg, ANNA_FILE_LOCATION);
  }
}


//------------------------------------------------------------------------------
//--------------------------------------------------------------- Avp::flagsOK()
//------------------------------------------------------------------------------
bool Avp::flagsOK() const throw() {
  // Dictionary stack avp:
  const stack::Avp *stackAvp = getStackAvp();

  if(!stackAvp) {
    std::string msg = "Impossible to decide if flags are correct because stack avp is not identified. Assume flags ok for Avp ";
    msg += anna::diameter::functions::avpIdAsPairString(a_id);
    anna::Logger::error(msg, ANNA_FILE_LOCATION);
    return true;
  };

  // Assignments
  bool Vnone = (stackAvp->getVbit() == stack::Avp::FlagRule::None);

  bool Vmust = (stackAvp->getVbit() == stack::Avp::FlagRule::must);

  bool Vmustnot = (stackAvp->getVbit() == stack::Avp::FlagRule::mustnot);

  bool Mnone = (stackAvp->getMbit() == stack::Avp::FlagRule::None);

  bool Mmust = (stackAvp->getMbit() == stack::Avp::FlagRule::must);

  bool Mmustnot = (stackAvp->getMbit() == stack::Avp::FlagRule::mustnot);

  bool Pnone = (stackAvp->getPbit() == stack::Avp::FlagRule::None);

  bool Pmust = (stackAvp->getPbit() == stack::Avp::FlagRule::must);

  bool Pmustnot = (stackAvp->getPbit() == stack::Avp::FlagRule::mustnot);

  // V-bit
  if((Vnone && Mnone && Pnone && vendorBit()) || (Vmust && !vendorBit()) || (Vmustnot && vendorBit())) {
    std::string msg = "Vendor Bit (V) incoherence found for Avp ";
    msg += stackAvp->getName();
    anna::Logger::error(msg, ANNA_FILE_LOCATION);
    return false;
  }

  // Exit on permissive mode:
  if(getEngine()->ignoreFlagsOnValidation()) return true;  // Ignore checking for AVP 'M' & 'P' bits

  // M-bit
  if((Mmust && !mandatoryBit()) || (Mmustnot && mandatoryBit())) {
    std::string msg = "Mandatory Bit (M) incoherence found for Avp ";
    msg += stackAvp->getName();
    anna::Logger::error(msg, ANNA_FILE_LOCATION);
    return false;
  }

  // P-bit
  if((Pmust && !encryptionBit()) || (Pmustnot && encryptionBit())) {
    std::string msg = "Encryption Bit (P) incoherence found for Avp ";
    msg += stackAvp->getName();
    anna::Logger::error(msg, ANNA_FILE_LOCATION);
    return false;
  }

  // Reserved bits
  if((a_flags & 0x1f) != 0x00) {
    std::string msg = "Any (or more than one) of the reserved avp flags bit has been activated. Reserved bits must be null. Avp is ";
    msg += stackAvp->getName();
    anna::Logger::error(msg, ANNA_FILE_LOCATION);
    return false;
  }

  return true;
}


//------------------------------------------------------------------------------
//----------------------------------------------------------------- Avp::setId()
//------------------------------------------------------------------------------
void Avp::setId(AvpId id) throw(anna::RuntimeException) {
  // Generic AVP assignment has no sense
  if(id == helpers::AVPID__AVP) return;

  // Clear class content:
  clear();
  // Id assignment:
  a_id = id;
  // Dictionary stack avp and format:
  const stack::Avp *stackAvp = getStackAvp(); // based on dictionary and a_id
  const stack::Format *stackFormat = stackAvp ? (stackAvp->getFormat()) : NULL /*Unknown*/;

  // Dictionary flags for known types:
  if(stackAvp) {
    if(stackAvp->getVbit() == stack::Avp::FlagRule::must) a_flags |= VBitMask;

    if(stackAvp->getMbit() == stack::Avp::FlagRule::must) a_flags |= MBitMask;

    if(stackAvp->getPbit() == stack::Avp::FlagRule::must) a_flags |= PBitMask;
  } else {
    if(a_id.second != 0) a_flags |= VBitMask; else a_flags &= (~VBitMask);
  }

  if(!stackFormat) {  // Unknown Avp
    LOGDEBUG(
      std::string msg = "Unknown format for AVP identifier ";
      msg += anna::diameter::functions::avpIdAsPairString(id);
      msg += ". Raw data without specific format will be managed";
      anna::Logger::debug(msg, ANNA_FILE_LOCATION);
    );
    a_Unknown = new Unknown();
    return;
  }

  // Avp class formats (Avp child classes should implement this source code block for application-specific formats):
  if(stackFormat->isOctetString()) a_OctetString = new OctetString();
  else if(stackFormat->isInteger32()) a_Integer32 = new Integer32();
  else if(stackFormat->isInteger64()) a_Integer64 = new Integer64();
  else if(stackFormat->isUnsigned32()) a_Unsigned32 = new Unsigned32();
  else if(stackFormat->isUnsigned64()) a_Unsigned64 = new Unsigned64();
  else if(stackFormat->isFloat32()) a_Float32 = new Float32();
  else if(stackFormat->isFloat64()) a_Float64 = new Float64();
  //else if (stackFormat->isGrouped())
  else if(stackFormat->isAddress()) a_Address = new Address();
  else if(stackFormat->isTime()) a_Time = new Time();
  else if(stackFormat->isUTF8String()) a_UTF8String = new UTF8String();
  else if(stackFormat->isDiameterIdentity()) a_DiameterIdentity = new DiameterIdentity();
  else if(stackFormat->isDiameterURI()) a_DiameterURI = new DiameterURI();
  else if(stackFormat->isEnumerated()) a_Enumerated = new Enumerated();
  else if(stackFormat->isIPFilterRule()) a_IPFilterRule = new IPFilterRule();
  else if(stackFormat->isQoSFilterRule()) a_QoSFilterRule = new QoSFilterRule();

  /////////////////////
  // Format specific //
  /////////////////////
  allocationByFormat(stackFormat);
}


//------------------------------------------------------------------------------
//----------------------------------------------------------------- Avp::setId()
//------------------------------------------------------------------------------
void Avp::setId(const char *name) throw(anna::RuntimeException) {
  setId(getEngine()->avpIdForName(name));
}


//------------------------------------------------------------------------------
//---------------------------------------------------------------- Avp::addAvp()
//------------------------------------------------------------------------------
Avp * Avp::addAvp(const char *name) throw(anna::RuntimeException) {
  return addAvp(getEngine()->avpIdForName(name));
}


//------------------------------------------------------------------------------
//------------------------------------------------------------- Avp::removeAvp()
//------------------------------------------------------------------------------
bool Avp::removeAvp(const char *name, int ocurrence) throw(anna::RuntimeException) {
  return removeAvp(getEngine()->avpIdForName(name), ocurrence);
}


//------------------------------------------------------------------------------
//--------------------------------------------------------------- Avp::_getAvp()
//------------------------------------------------------------------------------
const Avp *Avp::_getAvp(const char *name, int ocurrence, anna::Exception::Mode::_v emode) const throw(anna::RuntimeException) {
  return getAvp(getEngine()->avpIdForName(name), ocurrence, emode);
}


//------------------------------------------------------------------------------
//-------------------------------------------------------------- Avp::countAvp()
//------------------------------------------------------------------------------
int Avp::countAvp(const char *name) const throw(anna::RuntimeException) {
  return countAvp(getEngine()->avpIdForName(name));
}

//------------------------------------------------------------------------------
//------------------------------------------------------------- Avp::getLength()
//------------------------------------------------------------------------------
U24 Avp::getLength() const throw() {
  U24 result;
  // Avp format:
  const stack::Avp *stackAvp = getStackAvp();
  const stack::Format *stackFormat = stackAvp ? (stackAvp->getFormat()) : NULL /*Unknown*/;
  // Header length:
  result = vendorBit() ? HeaderLengthVactive : HeaderLengthVinactive;

  if(!stackFormat) {
    result += a_Unknown->getSize();
    return result;
  }

  if(stackFormat->isOctetString()) result += a_OctetString->getSize();
  else if(stackFormat->isInteger32()) result += a_Integer32->getSize();
  else if(stackFormat->isInteger64()) result += a_Integer64->getSize();
  else if(stackFormat->isUnsigned32()) result += a_Unsigned32->getSize();
  else if(stackFormat->isUnsigned64()) result += a_Unsigned64->getSize();
  else if(stackFormat->isFloat32()) result += a_Float32->getSize();
  else if(stackFormat->isFloat64()) result += a_Float64->getSize();
  else if(stackFormat->isGrouped()) for(const_avp_iterator it = avp_begin(); it != avp_end(); it++) result += 4 * REQUIRED_WORDS(avp(it)->getLength());
  else if(stackFormat->isAddress()) result += a_Address->getSize();
  else if(stackFormat->isTime()) result += a_Time->getSize();
  else if(stackFormat->isUTF8String()) result += a_UTF8String->getSize();
  else if(stackFormat->isDiameterIdentity()) result += a_DiameterIdentity->getSize();
  else if(stackFormat->isDiameterURI()) result += a_DiameterURI->getSize();
  else if(stackFormat->isEnumerated()) result += a_Enumerated->getSize();
  else if(stackFormat->isIPFilterRule()) result += a_IPFilterRule->getSize();
  else if(stackFormat->isQoSFilterRule()) result += a_QoSFilterRule->getSize();

  /////////////////////
  // Format specific //
  /////////////////////
  result += getLengthByFormat(stackFormat);
  return result;
}


//------------------------------------------------------------------------------
//-------------------------------------------- Avp::unknownAvpWithMandatoryBit()
//------------------------------------------------------------------------------
void Avp::unknownAvpWithMandatoryBit() const throw(anna::RuntimeException) {
  OamModule &oamModule = OamModule::instantiate();
  const char *c_aid = STRING_WITH_QUOTATION_MARKS__C_STR(anna::diameter::functions::avpIdAsPairString(a_id));
  oamModule.activateAlarm(OamModule::Alarm::AvpDecode__UnknownAvp__s__WithMandatoryBit, c_aid);
  oamModule.count(OamModule::Counter::AvpDecode__UnknownAvpWithMandatoryBit);
  LOGWARNING(
    std::string msg = anna::functions::asString("Detected unknown Avp %s with mandatory bit activated", c_aid);
    anna::Logger::warning(msg, ANNA_FILE_LOCATION);
  );
}

//------------------------------------------------------------------------------
//-------------------------------------------------------- Avp::decodeDataPart()
//------------------------------------------------------------------------------
void Avp::decodeDataPart(const char * buffer, int size, const parent_t & parent, Message *answer) throw(anna::RuntimeException) {
  // OAM
  OamModule &oamModule = OamModule::instantiate();
  // Dictionary stack avp and format:
  const stack::Avp *stackAvp = getStackAvp();
  const stack::Format *stackFormat = stackAvp ? (stackAvp->getFormat()) : NULL /*Unknown*/;

  if(!stackFormat) {
    a_Unknown->decode(buffer, size);
    return;
  }

  if(stackFormat->isOctetString()) a_OctetString->decode(buffer, size);
  else if(stackFormat->isInteger32()) a_Integer32->decode(buffer, size);
  else if(stackFormat->isInteger64()) a_Integer64->decode(buffer, size);
  else if(stackFormat->isUnsigned32()) a_Unsigned32->decode(buffer, size);
  else if(stackFormat->isUnsigned64()) a_Unsigned64->decode(buffer, size);
  else if(stackFormat->isFloat32()) a_Float32->decode(buffer, size);
  else if(stackFormat->isFloat64()) a_Float64->decode(buffer, size);
  else if(stackFormat->isGrouped()) {
    if(size == 0) {
      LOGDEBUG(anna::Logger::debug("Grouped Avp has empty data part", ANNA_FILE_LOCATION));
      return;
    }

    if((size % 4) != 0) {
      oamModule.activateAlarm(OamModule::Alarm::AvpDecode__IncorrectLength);
      oamModule.count(OamModule::Counter::AvpDecode__IncorrectLength);

      if(answer) {
        answer->setResultCode(helpers::base::AVPVALUES__Result_Code::DIAMETER_INVALID_AVP_LENGTH);
        answer->setFailedAvp(parent, a_id);
      }

      throw anna::RuntimeException("Avp format error, the avp length is incorrect (must be multiple of 4 on grouped type)", ANNA_FILE_LOCATION);
    }

    int avpPos = 0;
    Avp* avp;
    anna::DataBlock db;
    // Me as parent:
    parent_t me = parent;
    me.addAvp(a_id);

    while(avpPos < size) {
      try {
        avp = getEngine()->createAvp(NULL);
        db.assign(buffer + avpPos, size - avpPos /* is valid to pass total size (indeed i don't know the real avp size) because it will be limited and this has deep copy disabled (no memory is reserved) */);
        avp -> decode(db, me, answer);
      } catch(anna::RuntimeException &ex) {
        getEngine()->releaseAvp(avp);
        throw;
      }

      addChild(avp);
      avpPos += 4 * REQUIRED_WORDS(avp->getLength());
    }
  } else if(stackFormat->isAddress()) a_Address->decode(buffer, size);
  else if(stackFormat->isTime()) a_Time->decode(buffer, size);
  else if(stackFormat->isUTF8String()) a_UTF8String->decode(buffer, size);
  else if(stackFormat->isDiameterIdentity()) a_DiameterIdentity->decode(buffer, size);
  else if(stackFormat->isDiameterURI()) a_DiameterURI->decode(buffer, size);
  else if(stackFormat->isEnumerated()) a_Enumerated->decode(buffer, size);
  else if(stackFormat->isIPFilterRule()) a_IPFilterRule->decode(buffer, size);
  else if(stackFormat->isQoSFilterRule()) a_QoSFilterRule->decode(buffer, size);

  /////////////////////
  // Format specific //
  /////////////////////
  decodeDataPartByFormat(buffer, size, stackFormat);
}

//------------------------------------------------------------------------------
//---------------------------------------------------------------- Avp::decode()
//------------------------------------------------------------------------------
void Avp::decode(const anna::DataBlock &db, const parent_t & parent, Message *answer) throw(anna::RuntimeException) {
  // OAM
  OamModule &oamModule = OamModule::instantiate();

  if(db.getSize() < HeaderLengthVinactive) {
    oamModule.activateAlarm(OamModule::Alarm::AvpDecode__NotEnoughBytesToCoverAvpHeaderLength);
    oamModule.count(OamModule::Counter::AvpDecode__NotEnoughBytesToCoverAvpHeaderLength);
    // DIAMETER_INVALID_AVP_LENGTH; // no podr� construir un avp fiable, as� que no registro el result-code
    throw anna::RuntimeException("Not enough bytes to cover avp header length", ANNA_FILE_LOCATION);
  }

  const char * buffer = db.getData();

  //       0                   1                   2                   3
  //       0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
  //      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
  //      |                           AVP Code                            |
  //      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
  //      |   AVP Flags   |                  AVP Length                   |
  //      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
  //      |                        Vendor-ID (opt)                        |
  //      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
  //      |    Data ...
  //      +-+-+-+-+-+-+-+-+
  // Code flags and vendor-id
  U32 code = DECODE4BYTES_INDX_VALUETYPE(buffer, 0, U32);

  a_flags = (S8)buffer[4];

  U32 vendorID = helpers::VENDORID__base;

  if(vendorBit()) {
    if(db.getSize() < HeaderLengthVactive) {
      oamModule.activateAlarm(OamModule::Alarm::AvpDecode__NotEnoughBytesToCoverAvpHeaderLength);
      oamModule.count(OamModule::Counter::AvpDecode__NotEnoughBytesToCoverAvpHeaderLength);
      // DIAMETER_INVALID_AVP_LENGTH; // no podr� construir un avp fiable, as� que no registro el result-code
      throw anna::RuntimeException(anna::functions::asString("Not enough bytes to cover avp header length (avp code = %u)", code), ANNA_FILE_LOCATION);
    }

    vendorID = DECODE4BYTES_INDX_VALUETYPE(buffer, 8, U32);

    if(vendorID == helpers::VENDORID__base) {
      //      A vendor ID value of zero (0) corresponds to the IETF adopted AVP
      //      values, as managed by the IANA. Since the absence of the vendor
      //      ID field implies that the AVP in question is not vendor specific,
      //      ...
      //      implementations MUST NOT use the zero (0) vendor ID.
      oamModule.activateAlarm(OamModule::Alarm::AvpDecode__IncoherenceBetweenActivatedVBitAndZeroedVendorIDValueReceived);
      oamModule.count(OamModule::Counter::AvpDecode__IncoherenceBetweenActivatedVBitAndZeroedVendorIDValueReceived);
      throw anna::RuntimeException(anna::functions::asString("Incoherence between activated V bit and zeroed vendor-id value received (avp code = %u)", code), ANNA_FILE_LOCATION);
    }
  }

  // Avp identifier
  setId(AvpId(code, vendorID));
  // Flags could have been updated regarding dictionary, but during decoding we must respect buffer received:
  a_flags = (S8)buffer[4];

  // Here i know id and flags:
  //   The 'M' Bit, known as the Mandatory bit, indicates whether support of the AVP is required. If an AVP with the 'M' bit set is received by
  //   a Diameter client, server, proxy, or translation agent and either the AVP or its value is unrecognized, the message MUST be rejected.
  //   Diameter Relay and redirect agents MUST NOT reject messages with unrecognized AVPs.
  if(!getStackAvp() && mandatoryBit()) {
    if(answer) {
      answer->setResultCode(helpers::base::AVPVALUES__Result_Code::DIAMETER_AVP_UNSUPPORTED);
      answer->setFailedAvp(parent, a_id);
    }

    unknownAvpWithMandatoryBit();
  }

  // Avp Length
  U24 length = DECODE3BYTES_INDX_VALUETYPE(buffer, 5, U24);
  // Data start position:
  int startDataPos = vendorBit() ? HeaderLengthVactive : HeaderLengthVinactive;
  // Data part:
  int dataBytes = length - startDataPos;

  if(dataBytes < 0) {
    oamModule.activateAlarm(OamModule::Alarm::AvpDecode__IncorrectLength);
    oamModule.count(OamModule::Counter::AvpDecode__IncorrectLength);

    if(answer) {
      answer->setResultCode(helpers::base::AVPVALUES__Result_Code::DIAMETER_INVALID_AVP_LENGTH);
      answer->setFailedAvp(parent, a_id);
    }

    throw anna::RuntimeException(anna::functions::asString("Avp format error, the avp length is incorrect (avp code = %u)", code), ANNA_FILE_LOCATION);
  }

  try {
    decodeDataPart(buffer + startDataPos, dataBytes, parent, answer);
  } catch(anna::RuntimeException &ex) {
    oamModule.activateAlarm(OamModule::Alarm::AvpDecode__DataPartInconsistence);
    oamModule.count(OamModule::Counter::AvpDecode__DataPartInconsistence);

    if(answer) {
      answer->setResultCode(helpers::base::AVPVALUES__Result_Code::DIAMETER_INVALID_AVP_VALUE);
      answer->setFailedAvp(parent, a_id);
    }

    throw anna::RuntimeException(anna::functions::asString("Internal Avp decoding error (avp code = %u): %s", code, ex.getText().c_str()), ANNA_FILE_LOCATION);
  }
}


//------------------------------------------------------------------------------
//----------------------------------------------------------- Avp::getStackAvp()
//------------------------------------------------------------------------------
const anna::diameter::stack::Avp *Avp::getStackAvp(AvpId id, Engine *engine) throw() {
  const stack::Dictionary * dictionary = engine->getDictionary();
  return (dictionary ? (dictionary->getAvp(id)) : NULL);
}


//------------------------------------------------------------------------------
//------------------------------------------------------------------- Avp::fix()
//------------------------------------------------------------------------------
void Avp::fix(avp_container &avps, find_container &finds, int &insertionPositionForChilds, anna::diameter::stack::const_avprule_iterator ruleBegin, anna::diameter::stack::const_avprule_iterator ruleEnd) throw() {
  // Exit cases
  if(avps.size() == 0) return;  // empty content, nothing to fix

  if(ruleBegin == ruleEnd) return;  // no reference. Cannot fix

  // Copy reference container and clear internal map in order to build a fixed version:
  avp_container unfixedAvps;

  for(const_avp_iterator it = avps.begin(); it != avps.end(); it++)
    unfixedAvps[(*it).first] = (*it).second;

  avps.clear();
  // Cache system reset (fix probably will invalidate cached findings):
  finds.clear();
  insertionPositionForChilds = 0;
  avp_iterator avp_it;
  Avp * avp;

  for(anna::diameter::stack::const_avprule_iterator rule_it = ruleBegin; rule_it != ruleEnd; rule_it++) {
    while((avp_it = Avp::avp_find(unfixedAvps, (*rule_it).second.getId(), 1)) != unfixedAvps.end()) {
      avp = (*avp_it).second;
      addChild(avps, insertionPositionForChilds, avp);
      unfixedAvps.erase(avp_it); // processed
    }
  }

  // Add remaining avps:
  for(avp_it = unfixedAvps.begin(); avp_it != unfixedAvps.end(); avp_it++) {
    avp = (*avp_it).second;

    if(avp) addChild(avps, insertionPositionForChilds, avp);
  }

  // Fix grouped Avps:
  for(avp_iterator it = avps.begin(); it != avps.end(); it++) {
    avp = (*it).second;

    if(avp->hasChildren()) avp->fix();
  }
}


//------------------------------------------------------------------------------
//------------------------------------------------------------------- Avp::fix()
//------------------------------------------------------------------------------
void Avp::fix() throw() {
  // Dictionary stack avp:
  const stack::Avp *stackAvp = getStackAvp();
  const stack::Format *stackFormat = stackAvp ? (stackAvp->getFormat()) : NULL /*Unknown*/;

  if(!stackAvp) {
    //LOGDEBUG(anna::Logger::debug("No dictionary avp reference found. Cannot fix.", ANNA_FILE_LOCATION));
    return;
  }

  if(!stackFormat || !stackFormat->isGrouped()) {
    //LOGDEBUG(anna::Logger::debug("Avp is not grouped. Nothing to fix.", ANNA_FILE_LOCATION));
    return;
  }

  fix(a_avps, (find_container&)a_finds, a_insertionPositionForChilds, stackAvp->avprule_begin(), stackAvp->avprule_end());
}


//------------------------------------------------------------------------------
//------------------------------------------------------------ Avp::validLevel()
//------------------------------------------------------------------------------
bool Avp::validLevel(const avp_container &avps, anna::diameter::stack::const_avprule_iterator ruleBegin, anna::diameter::stack::const_avprule_iterator ruleEnd, Engine * engine, const parent_t & parent, Message *answer) throw(anna::RuntimeException) {
  bool result = true;
  // OAM
  OamModule &oamModule = OamModule::instantiate();
  ///////////
  // Fixed //
  ///////////
  // auxiliary
  AvpId id;
  const_avp_iterator avp_it = avps.begin();
  anna::diameter::stack::const_avprule_iterator rule_it;
  bool okFixed;

  for(rule_it = ruleBegin; rule_it != ruleEnd; rule_it++) {
    okFixed = true;

    if((*rule_it).second.isFixed()) {
      if(avps.size() == 0) break;  // cardinality checking will show this anomaly

      if(avp_it != avps.end()) {
        id = Avp::avp(avp_it)->getId();

        if(id != (*rule_it).second.getId()) {
          okFixed = false;
          // Missing fixed rule: %s
        } else if(Avp::avp_find(avps, id, 2) != avps.end()) {
          okFixed = false;
          // More than one fixed ocurrency for rule %s (*)
        }

        avp_it++;
        // (*) for (int k = 0; k < (*rule_it).second.getQualMax(); k++) avp_it++
      } else
        okFixed = false;

      if(!okFixed) {
        result = false;
        // OAM & Depth management
        oamModule.activateAlarm(OamModule::Alarm::LevelValidation__MissingFixedRule__s__Inside__s__, STRING_WITH_QUOTATION_MARKS__C_STR((*rule_it).second.asString(false /*ommit dots & pair*/)), STRING_WITH_QUOTATION_MARKS__C_STR(parent.asString()));
        oamModule.count(OamModule::Counter::LevelValidation__MissingFixedRule);

        if(answer) {
          answer->setResultCode(helpers::base::AVPVALUES__Result_Code::DIAMETER_MISSING_AVP);
          answer->setFailedAvp(parent, (*rule_it).second.getId());
        }

        engine->validationAnomaly(anna::functions::asString("Missing fixed rule %s inside %s", STRING_WITH_QUOTATION_MARKS__C_STR((*rule_it).second.asString(false /*ommit dots & pair*/)), STRING_WITH_QUOTATION_MARKS__C_STR(parent.asString())));
      }
    } else break; // finish fixed
  }

  /////////////////
  // Cardinality // Mandatory control is implicit here
  /////////////////
  anna::diameter::stack::const_avprule_iterator generic_rule_it = ruleEnd;
  int min, max, amount;

  for(rule_it = ruleBegin; rule_it != ruleEnd; rule_it++) {
    if((*rule_it).second.isAny()) { generic_rule_it = rule_it; continue; }  // generic Avp will be managed after

    id = (*rule_it).second.getId();
    min = (*rule_it).second.getQualMin();
    max = (*rule_it).second.getQualMax(); // -1 means infinite
    amount = Avp::countAvp(avps, id);

    if((amount < min) || ((max != -1) && (amount > max))) {
      // Failed rule %s for cardinality (found %d items)
      result = false;
      // OAM & Depth management
      oamModule.activateAlarm(OamModule::Alarm::LevelValidation__FailedRule__s__ForCardinality_Found__d__ItemsInside__s__, STRING_WITH_QUOTATION_MARKS__C_STR((*rule_it).second.asString(false /*ommit dots & pair*/)), amount, STRING_WITH_QUOTATION_MARKS__C_STR(parent.asString()));
      oamModule.count(OamModule::Counter::LevelValidation__FailedRuleForCardinality);

      if(amount < min) {
        oamModule.count(OamModule::Counter::LevelValidation__FailedRuleForCardinalityLessThanNeeded);

        if(answer) {
          answer->setResultCode(helpers::base::AVPVALUES__Result_Code::DIAMETER_MISSING_AVP);
          answer->setFailedAvp(parent, id);
        }
      } else {
        oamModule.count(OamModule::Counter::LevelValidation__FailedRuleForCardinalityMoreThanNeeded);

        if(answer) {
          answer->setResultCode(helpers::base::AVPVALUES__Result_Code::DIAMETER_AVP_OCCURS_TOO_MANY_TIMES);
          answer->setFailedAvp(parent, id);
        }
      }

      engine->validationAnomaly(anna::functions::asString("Failed rule %s for cardinality (found %d items inside %s)", STRING_WITH_QUOTATION_MARKS__C_STR((*rule_it).second.asString(false /*ommit dots & pair*/)), amount, STRING_WITH_QUOTATION_MARKS__C_STR(parent.asString())));
    }
  }

  bool haveGeneric = (generic_rule_it != ruleEnd);
  /////////////////
  // Generic AVP //
  /////////////////
  bool foundInRules;
  std::map < AvpId, int /* dummy */ > disregarded; // 'Disregardeds' are Avps not in rules list
  std::map < AvpId, int /* dummy */ >::const_iterator disregarded_it;

  for(avp_it = avps.begin(); avp_it != avps.end(); avp_it++) {
    id = Avp::avp(avp_it)->getId();
    // Find rule for the child:
    foundInRules = false;

    for(rule_it = ruleBegin; rule_it != ruleEnd; rule_it++) {
      if((*rule_it).second.getId() == id) {
        foundInRules = true;
        break;
      }
    }

    if(!foundInRules) {
      disregarded[id] = 0;
    }
  }

  // Generic AVP cardinality checkings
  int disregardeds = disregarded.size();

  if(haveGeneric) {
    min = (*generic_rule_it).second.getQualMin();
    max = (*generic_rule_it).second.getQualMax(); // -1 means infinite
    amount = disregardeds;

    if((amount < min) || ((max != -1) && (amount > max))) {
      // Failed Generic AVP rule %s for cardinality (found %d disregarded items inside %s)
      result = false;
      // OAM & Depth management
      oamModule.activateAlarm(OamModule::Alarm::LevelValidation__FailedGenericAvpRule__s__ForCardinality_Found__d__DisregardedItemsInside__s__, STRING_WITH_QUOTATION_MARKS__C_STR((*generic_rule_it).second.asString(false /*ommit dots & pair*/)), amount, STRING_WITH_QUOTATION_MARKS__C_STR(parent.asString()));
      oamModule.count(OamModule::Counter::LevelValidation__FailedGenericAvpRuleForCardinalityFoundDisregardedItem);

      if(answer) {
        answer->setResultCode(helpers::base::AVPVALUES__Result_Code::DIAMETER_AVP_NOT_ALLOWED);
        answer->setFailedAvp(parent, id); // NO SENSE... what to put ?
      }

      engine->validationAnomaly(anna::functions::asString("Failed Generic AVP rule %s for cardinality (found %d disregarded items inside %s)", STRING_WITH_QUOTATION_MARKS__C_STR((*generic_rule_it).second.asString(false /*ommit dots & pair*/)), amount, STRING_WITH_QUOTATION_MARKS__C_STR(parent.asString())));
    }
  } else if(disregardeds) {  // When Generic AVP missing, no disregarded Avps are allowed
    // Found %d disregarded items inside %s and Generic AVP was not specified
    result = false;
    // Build disregarded list as string (unknown as avp id pair):
    std::string s_disregardeds = "";
    const stack::Avp *stackAvp;

    for(disregarded_it = disregarded.begin(); disregarded_it != disregarded.end(); disregarded_it++) {
      id = (*disregarded_it).first;
      stackAvp = getStackAvp(id, engine);
      s_disregardeds += (stackAvp ? (stackAvp->getName()) : (anna::diameter::functions::avpIdAsPairString(id)));
      s_disregardeds += ", ";

      // We wouldn't know where are these disregarded, but...
      if(answer) {
        answer->setResultCode(helpers::base::AVPVALUES__Result_Code::DIAMETER_AVP_NOT_ALLOWED);
        answer->setFailedAvp(parent, id);
      }
    }

    s_disregardeds.erase(s_disregardeds.size() - 2, 2); // remove last ', '
    // OAM & Depth management
    oamModule.activateAlarm(OamModule::Alarm::LevelValidation__FoundDisregardedItemsInside__s__AndGenericAVPWasNotSpecified__s__, STRING_WITH_QUOTATION_MARKS__C_STR(parent.asString()), STRING_WITH_QUOTATION_MARKS__C_STR(s_disregardeds));
    oamModule.count(OamModule::Counter::LevelValidation__FoundDisregardedItemsAndGenericAVPWasNotSpecified);
    engine->validationAnomaly(anna::functions::asString("Found disregarded items inside %s and Generic AVP was not specified: %s", STRING_WITH_QUOTATION_MARKS__C_STR(parent.asString()), STRING_WITH_QUOTATION_MARKS__C_STR(s_disregardeds)));
  }

  return result;
}


//------------------------------------------------------------------------------
//----------------------------------------------------------------- Avp::valid()
//------------------------------------------------------------------------------
bool Avp::valid(const parent_t & parent, Message *answer) const throw(anna::RuntimeException) {
  // OAM
  OamModule &oamModule = OamModule::instantiate();
  // Dictionary stack avp:
  const stack::Avp *stackAvp = getStackAvp();
  const stack::Format *stackFormat = stackAvp ? (stackAvp->getFormat()) : NULL /*Unknown*/;

  if(!stackAvp) {
    // No dictionary avp reference found. Cannot validate
    return true; // perhaps a unknown Avp
  }

  // Me as parent:
  parent_t me = parent;
  me.addAvp(a_id, stackAvp->getName().c_str());

  if(!stackFormat) {
    // No format avp reference found. Cannot validate
    return true; // perhaps a unknown Avp
  }

  //////////////////////////////
  // Flags coherence checking //
  //////////////////////////////
  bool result = flagsOK();

  if(!result) {
    // OAM & Depth management
    oamModule.activateAlarm(OamModule::Alarm::AvpValidation__Avp__s__Flags__d__DoesNotFulfillTheDefinedFlagRules__s__, STRING_WITH_QUOTATION_MARKS__C_STR(me.asString()), (int)a_flags, STRING_WITH_QUOTATION_MARKS__C_STR(stackAvp->getFlagRulesDescription()));
    oamModule.count(OamModule::Counter::AvpValidation__AvpFlagsDoesNotFulfillTheDefinedFlagRules);

    if(answer) {
      answer->setResultCode(helpers::base::AVPVALUES__Result_Code::DIAMETER_INVALID_AVP_BITS);
      answer->setFailedAvp(parent, a_id, stackAvp->getName().c_str()); // RFC 6733 says nothing about Failed-AVP in this case...
    }

    getEngine()->validationAnomaly(anna::functions::asString("The AVP %s flags (%d) does not fulfill the defined flag rules: %s", STRING_WITH_QUOTATION_MARKS__C_STR(me.asString()), (int)a_flags, STRING_WITH_QUOTATION_MARKS__C_STR(stackAvp->getFlagRulesDescription())));
  }

  //////////////////////
  // Enumerated range //
  //////////////////////
  if(stackFormat->isEnumerated()) {
    if(!stackAvp->allowEnum(a_Enumerated->getValue())) {
      result = false;
      // OAM & Depth management
      oamModule.activateAlarm(OamModule::Alarm::AvpValidation__EnumeratedAvp__s__WithValue__d__DoesNotComplyRestriction__s__, STRING_WITH_QUOTATION_MARKS__C_STR(me.asString()), a_Enumerated->getValue(), stackAvp->getEnums());
      oamModule.count(OamModule::Counter::AvpValidation__EnumeratedAvpWithValueDoesNotComplyRestriction);

      if(answer) {
        answer->setResultCode(helpers::base::AVPVALUES__Result_Code::DIAMETER_INVALID_AVP_VALUE);
        answer->setFailedAvp(parent, a_id, stackAvp->getName().c_str());
      }

      getEngine()->validationAnomaly(anna::functions::asString("Enumerated AVP %s with value %d does not comply to restriction: %s", STRING_WITH_QUOTATION_MARKS__C_STR(me.asString()), a_Enumerated->getValue(), stackAvp->getEnums()));
    }
  }

  ////////////////////
  // Level checking //
  ////////////////////
  result = Avp::validLevel(a_avps, stackAvp->avprule_begin(), stackAvp->avprule_end(), getEngine(), me, answer) && result;

  ////////////////////////
  // Childrens checking //
  ////////////////////////
  if(a_id == helpers::base::AVPID__Failed_AVP) return result;  // DON'T VALIDATE CONTENT FOR Failed-AVP, because it won't be valid...

  for(const_avp_iterator it = avp_begin(); it != avp_end(); it++)
    result = ((*it).second->valid(me, answer)) && result;

  return result;
}


//------------------------------------------------------------------------------
//------------------------------------------------------------------ Avp::code()
//------------------------------------------------------------------------------
void Avp::code(char* buffer, int &size) const throw(anna::RuntimeException) {
  // Code
  buffer[0] = (S8)(a_id.first >> 24);
  buffer[1] = (S8)(a_id.first >> 16);
  buffer[2] = (S8)(a_id.first >> 8);
  buffer[3] = (S8) a_id.first;
  // Flags and length
  size = getLength();
  buffer[4] = (S8) a_flags;
  buffer[5] = (S8)(size >> 16);
  buffer[6] = (S8)(size >> 8);
  buffer[7] = (S8) size;

  // Vendor-id
  if(vendorBit()) {
    buffer[8] = (S8)(a_id.second >> 24);
    buffer[9] = (S8)(a_id.second >> 16);
    buffer[10] = (S8)(a_id.second >> 8);
    buffer[11] = (S8) a_id.second;
  }

  // Data start position:
  int startDataPos = vendorBit() ? HeaderLengthVactive : HeaderLengthVinactive;
  // Data part:
  S8 * dataPart = buffer + startDataPos;
  int dataBytes; // not used but could be useful for checking (size - startDataPos)

  if(startDataPos == size) {
    LOGDEBUG(anna::Logger::debug("There is no data part to encode", ANNA_FILE_LOCATION));
    return;
  }

  // Avp format:
  const stack::Avp *stackAvp = getStackAvp();
  const stack::Format *stackFormat = stackAvp ? (stackAvp->getFormat()) : NULL /*Unknown*/;

  if(!stackFormat) {
    a_Unknown->code(dataPart, dataBytes);
    return;
  }

  if(stackFormat->isOctetString()) a_OctetString->code(dataPart, dataBytes);
  else if(stackFormat->isInteger32()) a_Integer32->code(dataPart, dataBytes);
  else if(stackFormat->isInteger64()) a_Integer64->code(dataPart, dataBytes);
  else if(stackFormat->isUnsigned32()) a_Unsigned32->code(dataPart, dataBytes);
  else if(stackFormat->isUnsigned64()) a_Unsigned64->code(dataPart, dataBytes);
  else if(stackFormat->isFloat32()) a_Float32->code(dataPart, dataBytes);
  else if(stackFormat->isFloat64()) a_Float64->code(dataPart, dataBytes);
  else if(stackFormat->isGrouped()) {
//      // Each avp encoding will remain padding octets depending on format. In order to
//      //  code grouped content (each avp will be multiple of 4) we clean the entire buffer
//      //  to ensure padding (easier than custom-made for each format):
//      memset(dataPart, 0, size - startDataPos);
    char *dataPartGrouped = dataPart;
    int dataBytesGrouped;

    for(const_avp_iterator it = avp_begin(); it != avp_end(); it++) {
      avp(it)->code(dataPartGrouped, dataBytesGrouped);
      dataPartGrouped = dataPartGrouped + 4 * REQUIRED_WORDS(dataBytesGrouped);
    }
  } else if(stackFormat->isAddress()) a_Address->code(dataPart, dataBytes);
  else if(stackFormat->isTime()) a_Time->code(dataPart, dataBytes);
  else if(stackFormat->isUTF8String()) a_UTF8String->code(dataPart, dataBytes);
  else if(stackFormat->isDiameterIdentity()) a_DiameterIdentity->code(dataPart, dataBytes);
  else if(stackFormat->isDiameterURI()) a_DiameterURI->code(dataPart, dataBytes);
  else if(stackFormat->isEnumerated()) a_Enumerated->code(dataPart, dataBytes);
  else if(stackFormat->isIPFilterRule()) a_IPFilterRule->code(dataPart, dataBytes);
  else if(stackFormat->isQoSFilterRule()) a_QoSFilterRule->code(dataPart, dataBytes);

  /////////////////////
  // Format specific //
  /////////////////////
  codeByFormat(dataPart, stackFormat);
}


//------------------------------------------------------------------------------
//------------------------------------------------------------ Avp::getXMLdata()
//------------------------------------------------------------------------------
std::string Avp::getXMLdata(bool & isHex, const stack::Format *stackFormat) const throw() {
  std::string result = "";
  isHex = false;
  // Unknown
//   try {
//      if (!stackFormat) {
//         isHex = true;
//         return a_Unknown->asHexString();
//      }
//   } catch (anna::RuntimeException &ex) {
//      ex.trace();
//   }

  if(!stackFormat) {
    isHex = true;
    // Tricky situation: if you change the dictionary dynamically, and a previous formatted avp
    //  becomes unknown (the change consists in remove Avps basically), then this would get a core
    //  dump: a_Unknown = NULL. We are not going to protect that situation because it represents a
    //  implementation fault, and there are many points which could have similar bad behaviour
    //  (those where we access directly the a_Unknown pointer).
    // The best way to afford this is ... TODO:
    // Freeze dictionary after use from any resource (avp, message), setting a flag which deny
    //  any modification in such dictionary. The best way to do this is on engine configuration
    //  for Avp o Message, where we could  invoke something like getEngine()->getDictionary()->freeze()

    return a_Unknown->asHexString(); // asHexString for OctetString cannot launch exception
  }

  // Special case for Address: could launch exception if not printable
  try {
    if(stackFormat->isAddress()) return a_Address->asPrintableString();
  } catch(anna::RuntimeException&) {
    //ex.trace();
    isHex = true;
    return a_Address->asHexString();
  }

  try {
    if(stackFormat->isOctetString()) {
      isHex = true;
      return a_OctetString->asHexString();
    }

    // Non-hexadecimal
    if(stackFormat->isInteger32()) return a_Integer32->asPrintableString();

    if(stackFormat->isInteger64()) return a_Integer64->asPrintableString();

    if(stackFormat->isUnsigned32()) return a_Unsigned32->asPrintableString();

    if(stackFormat->isUnsigned64()) return a_Unsigned64->asPrintableString();

    if(stackFormat->isFloat32()) return a_Float32->asPrintableString();

    if(stackFormat->isFloat64()) return a_Float64->asPrintableString();

    //if (stackFormat->isAddress()) return a_Address->asPrintableString();

    if(stackFormat->isTime()) return a_Time->asPrintableString();

    if(stackFormat->isUTF8String()) return a_UTF8String->asPrintableString();

    if(stackFormat->isDiameterIdentity()) return a_DiameterIdentity->asPrintableString();

    if(stackFormat->isDiameterURI()) return a_DiameterURI->asPrintableString();

    if(stackFormat->isEnumerated()) return a_Enumerated->asPrintableString();

    if(stackFormat->isIPFilterRule()) return a_IPFilterRule->asPrintableString();

    if(stackFormat->isQoSFilterRule()) return a_QoSFilterRule->asPrintableString();

    /////////////////////
    // Format specific //
    /////////////////////
    return getXMLdataByFormat(isHex, stackFormat);
  } catch(anna::RuntimeException &ex) {
    ex.trace();
  }

  return result;
}


//------------------------------------------------------------------------------
//---------------------------------------------------------------- Avp::decode()
//------------------------------------------------------------------------------
void Avp::decode(const anna::DataBlock &db) throw(anna::RuntimeException) {
  parent_t parent;
  parent.setMessage(CommandId(0, false), "No-Parent");
  decode(db, parent, NULL);
}


//------------------------------------------------------------------------------
//--------------------------------------------------------------- Avp::fromXML()
//------------------------------------------------------------------------------
void Avp::fromXML(const anna::xml::Node* avpNode) throw(anna::RuntimeException) {
  // <!ATTLIST avp id CDATA #IMPLIED code CDATA #IMPLIED vendor-code CDATA #IMPLIED flags CDATA #IMPLIED data CDATA #IMPLIED hex-data CDATA #IMPLIED>
  const anna::xml::Attribute *name, *code, *vendorCode, *flags, *data, *hexData;
  name = avpNode->getAttribute("name", false /* no exception */);
  code = avpNode->getAttribute("code", false /* no exception */);
  vendorCode = avpNode->getAttribute("vendor-code", false /* no exception */);
  flags = avpNode->getAttribute("flags", false /* no exception */);
  data = avpNode->getAttribute("data", false /* no exception */);
  hexData = avpNode->getAttribute("hex-data", false /* no exception */);
  // Dictionary
  const stack::Dictionary * dictionary = getEngine()->getDictionary();
  const stack::Avp *stackAvp = NULL;
  const stack::Format *stackFormat = NULL;

  // Compact mode
  if(name) {
    if(!dictionary) {
      std::string msg = "Error processing avp <name '"; msg += name->getValue();
      msg += "'>: no dictionary available. Load one or use <'code' + 'flags' + 'vendorCode'> instead of <'name'>";
      throw anna::RuntimeException(msg, ANNA_FILE_LOCATION);
    }

    stackAvp = dictionary->getAvp(name->getValue());

    if(!stackAvp) {
      std::string msg = "Error processing avp <name '"; msg += name->getValue();
      msg += "'>: no avp found for this identifier at available dictionary";
      throw anna::RuntimeException(msg, ANNA_FILE_LOCATION);
    }

    stackFormat = stackAvp->getFormat();
  }

  // Check attributes exclusiveness
  if(name) {  // compact mode
    bool allowFlagsField = ( stackAvp->getVbit() == anna::diameter::stack::Avp::FlagRule::may || stackAvp->getVbit() == anna::diameter::stack::Avp::FlagRule::shouldnot
                             || stackAvp->getMbit() == anna::diameter::stack::Avp::FlagRule::may || stackAvp->getMbit() == anna::diameter::stack::Avp::FlagRule::shouldnot
                             /* umm, perhaps we could omit for bit P, whic is deprecated ... */
                             /* || stackAvp->getPbit() == anna::diameter::stack::Avp::FlagRule::may || stackAvp->getPbit() == anna::diameter::stack::Avp::FlagRule::shouldnot*/ );
    if(code || (flags && !allowFlagsField) || vendorCode) {
      std::string msg = "Error processing avp <name '"; msg += name->getValue();
      if (flags) msg += "'>: avp attributes <'code' + 'flags' + 'vendorCode'> are not allowed if <'name'> is provided (also flags is not permitted: no may, no shouldnot)";
      else msg += "'>: avp attributes <'code' + 'vendorCode'> are not allowed if <'name'> is provided";
      throw anna::RuntimeException(msg, ANNA_FILE_LOCATION);
    }

    setId(stackAvp->getId());

    if (flags && allowFlagsField) {
      // Flags check
      int i_aux = flags->getIntegerValue();

      if(i_aux < 0 || i_aux > 256) {
        std::string msg = "Error processing avp <flags '"; msg += flags->getValue();
        msg += "': out of range [0,256]";
        throw anna::RuntimeException(msg, ANNA_FILE_LOCATION);
      }

      a_flags = i_aux;
    }


  } else {
    if(!code || !flags || !vendorCode) {
      std::string s_code = code ? code->getValue() : "?";
      std::string s_flags = flags ? flags->getValue() : "?";
      std::string s_vendorCode = vendorCode ? vendorCode->getValue() : "?";
      std::string msg = "Error processing avp <code '"; msg += s_code;
      msg += "' + flags '"; msg += s_flags;
      msg += "' + vendorCode '"; msg += s_vendorCode;
      msg += "'>: all must be provided if <'name'> don't";
      throw anna::RuntimeException(msg, ANNA_FILE_LOCATION);
    }

    // Code check
    int i_aux = code->getIntegerValue();

    if(i_aux < 0) {
      std::string msg = "Error processing avp <code '"; msg += code->getValue();
      msg += "': negative values are not allowed";
      throw anna::RuntimeException(msg, ANNA_FILE_LOCATION);
    }

    U24 u_code = i_aux;
    // Flags check
    i_aux = flags->getIntegerValue();

    if(i_aux < 0 || i_aux > 256) {
      std::string msg = "Error processing avp <flags '"; msg += flags->getValue();
      msg += "': out of range [0,256]";
      throw anna::RuntimeException(msg, ANNA_FILE_LOCATION);
    }

    a_flags = i_aux;
    int flagsBCK = a_flags;
    // VendorCode checks
    i_aux = vendorCode->getIntegerValue();

    if(i_aux < 0) {
      std::string msg = "Error processing avp <vendorCode '"; msg += vendorCode->getValue();
      msg += "': negative values are not allowed";
      throw anna::RuntimeException(msg, ANNA_FILE_LOCATION);
    }

    U32 u_vendorCode = i_aux;
    bool vendorSpecific1 = (u_vendorCode != 0);
    bool vendorSpecific2 = (((a_flags) & VBitMask) != 0x00);

    if(vendorSpecific1 != vendorSpecific2) {
      std::string msg = "Error processing avp <vendorCode '"; msg += vendorCode->getValue();
      msg += "' and avp <flags '"; msg += flags->getValue();
      msg += "': incompatible vendor-specific properties";
      throw anna::RuntimeException(msg, ANNA_FILE_LOCATION);
    }

    // Final assignments
    setId(AvpId(u_code, u_vendorCode));
    // Flags could have been updated regarding dictionary, but during parsing we must respect xml file:
    a_flags = flagsBCK;
    stackAvp = getStackAvp();
    stackFormat = stackAvp ? stackAvp->getFormat() : NULL;
  }

  // Childrens
  // Check if grouped not confirmed
  anna::xml::Node::const_child_iterator it;
  anna::xml::Node::const_child_iterator minii = avpNode->child_begin();
  anna::xml::Node::const_child_iterator maxii = avpNode->child_end();
  const bool hasChildren(minii != maxii);

  if(hasChildren) {
    // We need to confirm is grouped
    if(!stackFormat)
      throw anna::RuntimeException("An xml avp node with children nodes could not confirmed as Grouped", ANNA_FILE_LOCATION);

    if(!stackFormat->isGrouped())
      throw anna::RuntimeException("An xml avp node with children nodes is not grouped type", ANNA_FILE_LOCATION);

    if(data || hexData)
      throw anna::RuntimeException("An xml avp node with children nodes has 'data' or 'hex-data' field", ANNA_FILE_LOCATION);
  }

  // Presentation
  std::string s_avp = name ? (name->getValue()) : (anna::diameter::functions::avpIdAsPairString(a_id));

  if(data && hexData) {
    std::string msg = "Not allowed both 'data' and 'hex-data' definitions. Avp ";
    msg += s_avp;
    throw anna::RuntimeException(msg, ANNA_FILE_LOCATION);
  }

  LOGWARNING(
    bool unknown = !stackFormat;
    bool octetString = stackFormat && (stackFormat->isOctetString());

  if(data && (unknown || octetString)) {
  std::string msg = "Not recommended 'data' field at ";
  msg += unknown ? "Unknown" : "OctetString";
  msg += "-format Avp "; msg += s_avp;
  msg += ". Use better 'hex-data' or omit content fields if empty";
  anna::Logger::warning(msg, ANNA_FILE_LOCATION);
  }
  );

  if(!stackFormat) {
    if(data) a_Unknown->fromPrintableString(data->getValue().c_str());
    else if(hexData) a_Unknown->fromHexString(hexData->getValue());

    return;
  }

  if(stackFormat->isOctetString()) {
    if(data) a_OctetString->fromPrintableString(data->getValue().c_str());
    else if(hexData) a_OctetString->fromHexString(hexData->getValue());
  } else if(stackFormat->isInteger32()) {
    if(data) a_Integer32->fromPrintableString(data->getValue().c_str());
    else if(hexData) a_Integer32->fromHexString(hexData->getValue());
  } else if(stackFormat->isInteger64()) {
    if(data) a_Integer64->fromPrintableString(data->getValue().c_str());
    else if(hexData) a_Integer64->fromHexString(hexData->getValue());
  } else if(stackFormat->isUnsigned32()) {
    if(data) a_Unsigned32->fromPrintableString(data->getValue().c_str());
    else if(hexData) a_Unsigned32->fromHexString(hexData->getValue());
  } else if(stackFormat->isUnsigned64()) {
    if(data) a_Unsigned64->fromPrintableString(data->getValue().c_str());
    else if(hexData) a_Unsigned64->fromHexString(hexData->getValue());
  } else if(stackFormat->isFloat32()) {
    if(data) a_Float32->fromPrintableString(data->getValue().c_str());
    else if(hexData) a_Float32->fromHexString(hexData->getValue());
  } else if(stackFormat->isFloat64()) {
    if(data) a_Float64->fromPrintableString(data->getValue().c_str());
    else if(hexData) a_Float64->fromHexString(hexData->getValue());
  } else if(stackFormat->isGrouped()) {
    // Childrens
    Avp *avp;

    for(it = minii; it != maxii; it++) {
      std::string nodeName = (*it)->getName();

      if(nodeName != "avp") {
        std::string msg = "Unsupported grouped avp child node name '"; msg += nodeName;
        msg += "': use 'avp'";
        throw anna::RuntimeException(msg, ANNA_FILE_LOCATION);
      }

      try {
        avp =  getEngine()->createAvp(NULL);
        avp -> fromXML(*it);
      } catch(anna::RuntimeException &ex) {
        getEngine()->releaseAvp(avp);
        throw;
      }

      addAvp(avp);
    }
  } else if(stackFormat->isAddress()) {
    if(data) a_Address->fromPrintableString(data->getValue().c_str());
    else if(hexData) a_Address->fromHexString(hexData->getValue());
  } else if(stackFormat->isTime()) {
    if(data) a_Time->fromPrintableString(data->getValue().c_str());
    else if(hexData) a_Time->fromHexString(hexData->getValue());
  } else if(stackFormat->isUTF8String()) {
    if(data) a_UTF8String->fromPrintableString(data->getValue().c_str());
    else if(hexData) a_UTF8String->fromHexString(hexData->getValue());
  } else if(stackFormat->isDiameterIdentity()) {
    if(data) a_DiameterIdentity->fromPrintableString(data->getValue().c_str());
    else if(hexData) a_DiameterIdentity->fromHexString(hexData->getValue());
  } else if(stackFormat->isDiameterURI()) {
    if(data) a_DiameterURI->fromPrintableString(data->getValue().c_str());
    else if(hexData) a_DiameterURI->fromHexString(hexData->getValue());
  } else if(stackFormat->isEnumerated()) {
    if(data) a_Enumerated->fromPrintableString(data->getValue().c_str());
    else if(hexData) a_Enumerated->fromHexString(hexData->getValue());
  } else if(stackFormat->isIPFilterRule()) {
    if(data) a_IPFilterRule->fromPrintableString(data->getValue().c_str());
    else if(hexData) a_IPFilterRule->fromHexString(hexData->getValue());
  } else if(stackFormat->isQoSFilterRule()) {
    if(data) a_QoSFilterRule->fromPrintableString(data->getValue().c_str());
    else if(hexData) a_QoSFilterRule->fromHexString(hexData->getValue());
  }

  /////////////////////
  // Format specific //
  /////////////////////
  fromXMLByFormat(data, hexData, stackFormat);
}

//------------------------------------------------------------------------------
//----------------------------------------------------------------- Avp::asXML()
//------------------------------------------------------------------------------
anna::xml::Node* Avp::asXML(anna::xml::Node* parent) const throw() {
  // <!ATTLIST avp id CDATA #IMPLIED code CDATA #IMPLIED vendor-code CDATA #IMPLIED flags CDATA #IMPLIED data CDATA #IMPLIED hex-data CDATA #IMPLIED>
  anna::xml::Node* result = parent->createChild("avp");
  // Dictionary stack avp and format:
  const stack::Avp *stackAvp = getStackAvp();
  const stack::Format *stackFormat = stackAvp ? (stackAvp->getFormat()) : NULL /*Unknown*/;
  bool compactMode = (stackAvp && flagsOK());

  if(compactMode) {
    result->createAttribute("name", stackAvp->getName());
    // If may or shouldnot is present in AVP definition, we have to show flags to avoid uncertainty
    if ( stackAvp->getVbit() == anna::diameter::stack::Avp::FlagRule::may || stackAvp->getVbit() == anna::diameter::stack::Avp::FlagRule::shouldnot
         || stackAvp->getMbit() == anna::diameter::stack::Avp::FlagRule::may || stackAvp->getMbit() == anna::diameter::stack::Avp::FlagRule::shouldnot
         /* umm, perhaps we could omit for bit P, whic is deprecated ... */
         /* || stackAvp->getPbit() == anna::diameter::stack::Avp::FlagRule::may || stackAvp->getPbit() == anna::diameter::stack::Avp::FlagRule::shouldnot*/ )
      result->createAttribute("flags", (int)a_flags);
  } else {
    result->createAttribute("code", a_id.first);
    result->createAttribute("vendor-code", a_id.second);
    result->createAttribute("flags", (int)a_flags);
  }

  // Grouped special case:
  if(stackFormat && stackFormat->isGrouped()) {
    for(const_avp_iterator it = avp_begin(); it != avp_end(); it++) avp(it)->asXML(result);

    return result;
  }

  // Data part literals:
  bool isHex;
  std::string value = getXMLdata(isHex, stackFormat);

  if(isHex) {
    result->createAttribute("hex-data", value);
  } else {
    result->createAttribute("data", value);
    const char *alias = stackAvp->getAlias(value);

    if(alias) result->createAttribute("alias", alias);  // additional information
  }

  return result;
}


//------------------------------------------------------------------------------
//----------------------------------------------------------- Avp::asXMLString()
//------------------------------------------------------------------------------
std::string Avp::asXMLString() const throw() {
  anna::xml::Node root("root");
  return anna::xml::Compiler().apply(asXML(&root));
}


//------------------------------------------------------------------------------
//---------------------------------------------------------------- Avp::isLike()
//------------------------------------------------------------------------------
bool Avp::isLike(const std::string &pattern) const throw() {
  anna::RegularExpression re(pattern);
  return re.isLike(asXMLString());
}
